"""Utils for Faster RCNN estimator"""
import os

from mxnet import gluon

from ....data.batchify import FasterRCNNTrainBatchify, Tuple, Append
from ....data.sampler import SplitSortedBucketSampler
from ....data.transforms.presets.rcnn import FasterRCNNDefaultValTransform
from .... import data as gdata
from ....utils.metrics.coco_detection import COCODetectionMetric
from ....utils.metrics.voc_detection import VOC07MApMetric

try:
    import horovod.mxnet as hvd
except ImportError:
    hvd = None


def _get_lr_at_iter(alpha, lr_warmup_factor=1. / 3.):
    return lr_warmup_factor * (1 - alpha) + alpha


def _split_and_load(batch, ctx_list):
    """Split data to 1 batch each device."""
    new_batch = []
    for _, data in enumerate(batch):
        if isinstance(data, (list, tuple)):
            new_data = [x.as_in_context(ctx) for x, ctx in zip(data, ctx_list)]
        else:
            new_data = [data.as_in_context(ctx_list[0])]
        new_batch.append(new_data)
    return new_batch


def _save_params(net, logger, best_map, current_map, epoch, save_interval, prefix):
    current_map = float(current_map)
    if current_map > best_map[0]:
        logger.info('[Epoch {}] mAP {} higher than current best {} saving to {}'.format(
            epoch, current_map, best_map, '{:s}_best.params'.format(prefix)))
        best_map[0] = current_map
        net.save_parameters('{:s}_best.params'.format(prefix))
        with open(prefix + '_best_map.log', 'a') as log_file:
            log_file.write('{:04d}:\t{:.4f}\n'.format(epoch, current_map))
    if save_interval and (epoch + 1) % save_interval == 0:
        logger.info('[Epoch {}] Saving parameters to {}'.format(
            epoch, '{:s}_{:04d}_{:.4f}.params'.format(prefix, epoch, current_map)))
        net.save_parameters('{:s}_{:04d}_{:.4f}.params'.format(prefix, epoch, current_map))


def _get_dataloader(net, train_dataset, val_dataset, train_transform, val_transform, batch_size,
                    num_shards, args):
    """Get dataloader."""
    train_bfn = FasterRCNNTrainBatchify(net, num_shards)
    if hasattr(train_dataset, 'get_im_aspect_ratio'):
        im_aspect_ratio = train_dataset.get_im_aspect_ratio()
    else:
        im_aspect_ratio = [1.] * len(train_dataset)
    train_sampler = \
        SplitSortedBucketSampler(im_aspect_ratio, batch_size,
                                 num_parts=hvd.size() if args.horovod else 1,
                                 part_index=hvd.rank() if args.horovod else 0,
                                 shuffle=True)
    train_loader = gluon.data.DataLoader(
        train_dataset.transform(
            train_transform(net.short, net.max_size, net, ashape=net.ashape,
                            multi_stage=args.faster_rcnn.use_fpn)),
        batch_sampler=train_sampler, batchify_fn=train_bfn, num_workers=args.num_workers)
    val_bfn = Tuple(*[Append() for _ in range(3)])
    short = net.short[-1] if isinstance(net.short, (tuple, list)) else net.short
    # validation use 1 sample per device
    val_loader = gluon.data.DataLoader(
        val_dataset.transform(val_transform(short, net.max_size)), num_shards, False,
        batchify_fn=val_bfn, last_batch='keep', num_workers=args.num_workers)
    train_eval_loader = gluon.data.DataLoader(
        train_dataset.transform(val_transform(short, net.max_size)), num_shards, False,
        batchify_fn=val_bfn, last_batch='keep', num_workers=args.num_workers)
    return train_loader, val_loader, train_eval_loader


def _get_testloader(net, test_dataset, num_devices, config):
    """Get faster rcnn test dataloader."""
    if config.meta_arch == 'faster_rcnn':
        test_bfn = Tuple(*[Append() for _ in range(3)])
        short = net.short[-1] if isinstance(net.short, (tuple, list)) else net.short
        # validation use 1 sample per device
        test_loader = gluon.data.DataLoader(
            test_dataset.transform(FasterRCNNDefaultValTransform(short, net.max_size)),
            num_devices,
            False,
            batchify_fn=test_bfn,
            last_batch='keep',
            num_workers=config.num_workers
        )
        return test_loader
    else:
        raise NotImplementedError('%s not implemented.' % config.meta_arch)


def _get_dataset(dataset, args):
    if dataset.lower() == 'voc':
        train_dataset = gdata.VOCDetection(
            splits=[(2007, 'trainval'), (2012, 'trainval')])
        val_dataset = gdata.VOCDetection(
            splits=[(2007, 'test')])
        val_metric = VOC07MApMetric(iou_thresh=0.5, class_names=val_dataset.classes)
    elif dataset.lower() == 'voc_tiny':
        # need to download the dataset and specify the path to store the dataset in
        # root = os.path.expanduser('~/.mxnet/datasets/')
        # filename_zip = ag.download('https://autogluon.s3.amazonaws.com/datasets/tiny_motorbike.zip', path=root)
        # filename = ag.unzip(filename_zip, root=root)
        # data_root = os.path.join(root, filename)
        train_dataset = gdata.CustomVOCDetectionBase(classes=('motorbike',), root=args.dataset_root + 'tiny_motorbike',
                                                     splits=[('', 'trainval')])
        val_dataset = gdata.CustomVOCDetectionBase(classes=('motorbike',), root=args.dataset_root + 'tiny_motorbike',
                                                   splits=[('', 'test')])
        val_metric = VOC07MApMetric(iou_thresh=0.5, class_names=val_dataset.classes)
    elif dataset.lower() in ['clipart', 'comic', 'watercolor']:
        root = os.path.join('~', '.mxnet', 'datasets', dataset.lower())
        train_dataset = gdata.CustomVOCDetection(root=root, splits=[('', 'train')],
                                                 generate_classes=True)
        val_dataset = gdata.CustomVOCDetection(root=root, splits=[('', 'test')],
                                               generate_classes=True)
        val_metric = VOC07MApMetric(iou_thresh=0.5, class_names=val_dataset.classes)
    elif dataset.lower() == 'coco':
        train_dataset = gdata.COCODetection(splits='instances_train2017', use_crowd=False)
        val_dataset = gdata.COCODetection(splits='instances_val2017', skip_empty=False)
        val_metric = COCODetectionMetric(val_dataset,
                                         os.path.join(args.logdir, args.save_prefix + '_eval'),
                                         cleanup=True)
    else:
        raise NotImplementedError('Dataset: {} not implemented.'.format(dataset))
    if args.train.mixup:
        from gluoncv.data.mixup import detection
        train_dataset = detection.MixupDetection(train_dataset)
    return train_dataset, val_dataset, val_metric



"""Base Estimator"""
import os
import pickle
import io
import logging
from datetime import datetime
import numpy as np
import pandas as pd
from ...utils import random as _random
from ...utils.filesystem import temporary_filename

logging.basicConfig(level=logging.INFO)

def set_default(cfg):
    """A special hook to register the default values for decorated Estimator.

    Parameters
    ----------
    cfg : autocfg.dataclass
        Default config as dataclass object
    """
    def _apply(cls):
        # docstring
        cls.__doc__ = str(cls.__doc__) if cls.__doc__ else ''
        cls.__doc__ += ("\n\nParameters\n"
                        "----------\n"
                        "config : str, dict\n"
                        "  Config used to override default configurations. \n"
                        "  If `str`, assume config file (.yml, .yaml) is used. \n"
                        "logger : logger, default is `None`.\n"
                        "  If not `None`, will use default logging object.\n"
                        "logdir : str, default is None.\n"
                        "  Directory for saving logs. If `None`, current working directory is used.\n")
        cls.__doc__ += '\nDefault configurations: \n--------------------\n'
        sio = io.StringIO()
        cfg.save(sio)
        cls.__doc__ += '\n' + sio.getvalue()
        cls._default_cfg = cfg
        cls._default_cfg.freeze()
        return cls
    return _apply


class BaseEstimator:
    """This is the base estimator for gluoncv.auto.Estimators.

    Parameters
    ----------
    config : dict
        Config in nested dict.
    logger : logging.Logger
        Optional logger for this estimator, can be `None` when default setting is used.
    reporter : callable
        The reporter for metric checkpointing.
    name : str
        Optional name for the estimator.

    Attributes
    ----------
    _logger : logging.Logger
        The customized/default logger for this estimator.
    _logdir : str
        The temporary dir for logs.
    _cfg : autocfg.dataclass
        The configurations.

    """
    def __init__(self, config, logger=None, reporter=None, name=None):
        # self._init_args = [config, logger, reporter]
        self._reporter = reporter
        name = name if isinstance(name, str) else self.__class__.__name__
        self._name = name
        self._logger = logger if logger is not None else logging.getLogger(name)
        self._logger.setLevel(logging.INFO)

        # reserved attributes
        self.net = None
        self.num_class = None
        self.classes = []
        self.ctx = [None]
        self.dataset = 'auto'

        # logdir
        logdir = config.pop('logdir', None)
        self._logdir = os.path.abspath(logdir) if logdir else os.getcwd()

        # finalize config
        cfg = self._default_cfg.merge(config)
        diffs = self._default_cfg.diff(cfg)
        if diffs:
            self._logger.info('modified configs: {')
            for diff in diffs:
                self._logger.info(diff)
            self._logger.info('}')
        self._cfg = cfg

        prefix = name.lower() + datetime.now().strftime("-%m-%d-%Y")
        self._logdir = os.path.join(self._logdir, prefix)
        # r.config['logdir'] = self._logdir
        os.makedirs(self._logdir, exist_ok=True)
        config_file = os.path.join(self._logdir, 'config.yaml')
        # log file
        self._log_file = os.path.join(self._logdir, 'estimator.log')
        fh = logging.FileHandler(self._log_file)
        self._logger.addHandler(fh)
        # save_config(r.config, self._logger, config_file)
        self._cfg.save(config_file)
        self._logger.info('Saved config to %s', config_file)

        # freeze config
        self._cfg.freeze()
        seed = self._cfg.get('seed', np.random.randint(1000000))
        _random.seed(seed)

    def fit(self, train_data, val_data=None, train_size=0.9, random_state=None, resume=False):
        """Fit with train/validation data.

        Parameters
        ----------
        train_data : pd.DataFrame or iterator
            Training data.
        val_data : pd.DataFrame or iterator, optional
            Validation data, optional. If `train_data` is DataFrame, `val_data` will be split from
            `train_data` given `train_size`.
        train_size : float
            The portion of train data split from original `train_data` if `val_data` is not provided.
        random_state : int
            Random state for splitting, for `np.random.seed`.
        resume : bool
            Whether resume from previous `fit`(if possible) or start as fresh.

        Returns
        -------
        None

        """
        if not resume:
            self.classes = train_data.classes
            self.num_class = len(self.classes)
            self._init_network()
        if not isinstance(train_data, pd.DataFrame):
            assert val_data is not None, \
                "Please provide `val_data` as we do not know how to split `train_data` of type: \
                {}".format(type(train_data))
            return self._fit(train_data, val_data) if not resume else self._resume_fit(train_data, val_data)

        if val_data is None:
            assert 0 <= train_size <= 1.0
            if random_state:
                np.random.seed(random_state)
            split_mask = np.random.rand(len(train_data)) < train_size
            train = train_data[split_mask]
            val = train_data[~split_mask]
            self._logger.info('Randomly split train_data into train[%d]/validation[%d] splits.',
                              len(train), len(val))
            return self._fit(train, val) if not resume else self._resume_fit(train, val)

        return self._fit(train_data, val_data) if not resume else self._resume_fit(train_data, val_data)

    def evaluate(self, val_data):
        """Evaluate estimator on validation data.

        Parameters
        ----------
        val_data : pd.DataFrame or iterator
            The validation data.

        """
        return self._evaluate(val_data)

    def predict(self, x):
        """Predict using this estimator.

        Parameters
        ----------
        x : str, pd.DataFrame or ndarray
            The input, can be str(filepath), pd.DataFrame with 'image' column, or raw ndarray input.

        """
        return self._predict(x)

    def predict_feature(self, x):
        """Predict intermediate features using this estimator.

        Parameters
        ----------
        x : str, pd.DataFrame or ndarray
            The input, can be str(filepath), pd.DataFrame with 'image' column, or raw ndarray input.

        """
        return self._predict_feature(x)

    def _predict(self, x):
        raise NotImplementedError

    def _predict_feature(self, x):
        raise NotImplementedError

    def _fit(self, train_data, val_data):
        raise NotImplementedError

    def _resume_fit(self, train_data, val_data):
        raise NotImplementedError

    def _evaluate(self, val_data):
        raise NotImplementedError

    def _init_network(self):
        raise NotImplementedError

    def _init_trainer(self):
        raise NotImplementedError

    def save(self, filename):
        """Save the state of this estimator to disk.

        Parameters
        ----------
        filename : str
            The file name for storing the full state.
        """
        with open(filename, 'wb') as fid:
            pickle.dump(self, fid)
        self._logger.info('Pickled to %s', filename)

    @classmethod
    def load(cls, filename):
        """Load the state from disk copy.

        Parameters
        ----------
        filename : str
            The file name to load from.
        """
        with open(filename, 'rb') as fid:
            obj = pickle.load(fid)
            obj._logger.info('Unpickled from %s', filename)
            return obj

    def __getstate__(self):
        d = self.__dict__.copy()
        try:
            import mxnet as mx
            d.pop('async_net', None)
            d.pop('_feature_net', None)
            net = d.get('net', None)
            if isinstance(net, mx.gluon.HybridBlock):
                with temporary_filename() as tfile:
                    net.save_parameters(tfile)
                    with open(tfile, 'rb') as fi:
                        d['net'] = fi.read()
            trainer = d.get('trainer', None)
            if isinstance(trainer, mx.gluon.Trainer):
                with temporary_filename() as tfile:
                    trainer.save_states(tfile)
                    with open(tfile, 'rb') as fi:
                        d['trainer'] = fi.read()
        except ImportError:
            pass
        d['_logger'] = None
        d['_reporter'] = None
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)
        # logger
        self._logger = logging.getLogger(state.get('_name', self.__class__.__name__))
        self._logger.setLevel(logging.INFO)
        fh = logging.FileHandler(self._log_file)
        self._logger.addHandler(fh)
        try:
            import mxnet as _
            net_params = state['net']
            self._init_network()
            with temporary_filename() as tfile:
                with open(tfile, 'wb') as fo:
                    fo.write(net_params)
                self.net.load_parameters(tfile)
            trainer_state = state['trainer']
            self._init_trainer()
            with temporary_filename() as tfile:
                with open(tfile, 'wb') as fo:
                    fo.write(trainer_state)
                self.trainer.load_states(tfile)
        except ImportError:
            pass



import numpy as np

cimport numpy as np
cimport cython

from libc.math cimport log

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def np_normalized_box_encoder(np.ndarray[float, ndim=2] samples, np.ndarray[long, ndim=2] matches,
                              np.ndarray[float, ndim=3] anchors, np.ndarray[float, ndim=3] refs,
                              np.ndarray[float, ndim=1] means, np.ndarray[float, ndim=1] stds):
    cdef int b = refs.shape[0]
    cdef int n = anchors.shape[1]
    cdef float ref_width, ref_height, ref_x, ref_y, ref_xmin, ref_ymin
    cdef float a_width, a_height, a_x, a_y, a_xmin, a_ymin
    cdef float t0, t1, t2, t3
    cdef int i, j, k, match
    cdef np.ndarray[float, ndim=3] targets = np.zeros((b, n, 4), dtype=np.float32)
    cdef np.ndarray[float, ndim=3] masks = np.zeros((b, n, 4), dtype=np.float32)
    with nogil:
        for i in range(b):
            for j in range(n):
                match = matches[i, j]
                # xmin: 0, ymin:1, xmax: 2, ymax: 3
                # x:0, y:1, w:2, h:3
                ref_xmin = refs[i, match, 0]
                ref_ymin = refs[i, match, 1]
                ref_width = refs[i, match, 2] - ref_xmin
                ref_height = refs[i, match, 3] - ref_ymin
                ref_x = ref_xmin + ref_width * 0.5
                ref_y = ref_ymin + ref_height * 0.5
                a_xmin = anchors[i, j, 0]
                a_ymin = anchors[i, j, 1]
                a_width = anchors[i, j, 2] - a_xmin
                a_height = anchors[i, j, 3] - a_ymin
                a_x = a_xmin + a_width * 0.5
                a_y = a_ymin + a_height * 0.5
                t0 = ((ref_x - a_x) / a_width - means[0]) / stds[0]
                t1 = ((ref_y - a_y) / a_height - means[1]) / stds[1]
                t2 = (log(ref_width / a_width) - means[2]) / stds[2]
                t3 = (log(ref_height / a_height) - means[3]) / stds[3]
                valid = 1 if samples[i, j] > 0.5 else 0
                masks[i, j, 0] = valid
                masks[i, j, 1] = valid
                masks[i, j, 2] = valid
                masks[i, j, 3] = valid
                targets[i, j, 0] = t0 if valid else 0.0
                targets[i, j, 1] = t1 if valid else 0.0
                targets[i, j, 2] = t2 if valid else 0.0
                targets[i, j, 3] = t3 if valid else 0.0
    return targets, masks



"""CenterNet training target generator."""
from __future__ import absolute_import

import numpy as np

from mxnet import nd
from mxnet import gluon


class CenterNetTargetGenerator(gluon.Block):
    """Target generator for CenterNet.

    Parameters
    ----------
    num_class : int
        Number of categories.
    output_width : int
        Width of the network output.
    output_height : int
        Height of the network output.

    """
    def __init__(self, num_class, output_width, output_height):
        super(CenterNetTargetGenerator, self).__init__()
        self._num_class = num_class
        self._output_width = int(output_width)
        self._output_height = int(output_height)

    def forward(self, gt_boxes, gt_ids):
        """Target generation"""
        # pylint: disable=arguments-differ
        h_scale = 1.0  # already scaled in affinetransform
        w_scale = 1.0  # already scaled in affinetransform
        heatmap = np.zeros((self._num_class, self._output_height, self._output_width),
                           dtype=np.float32)
        wh_target = np.zeros((2, self._output_height, self._output_width), dtype=np.float32)
        wh_mask = np.zeros((2, self._output_height, self._output_width), dtype=np.float32)
        center_reg = np.zeros((2, self._output_height, self._output_width), dtype=np.float32)
        center_reg_mask = np.zeros((2, self._output_height, self._output_width), dtype=np.float32)
        for bbox, cid in zip(gt_boxes, gt_ids):
            cid = int(cid)
            box_h, box_w = bbox[3] - bbox[1], bbox[2] - bbox[0]
            if box_h > 0 and box_w > 0:
                radius = _gaussian_radius((np.ceil(box_h), np.ceil(box_w)))
                radius = max(0, int(radius))
                center = np.array(
                    [(bbox[0] + bbox[2]) / 2 * w_scale, (bbox[1] + bbox[3]) / 2 * h_scale],
                    dtype=np.float32)
                center_int = center.astype(np.int32)
                center_x, center_y = center_int
                assert center_x < self._output_width, \
                    'center_x: {} > output_width: {}'.format(center_x, self._output_width)
                assert center_y < self._output_height, \
                    'center_y: {} > output_height: {}'.format(center_y, self._output_height)
                _draw_umich_gaussian(heatmap[cid], center_int, radius)
                wh_target[0, center_y, center_x] = box_w * w_scale
                wh_target[1, center_y, center_x] = box_h * h_scale
                wh_mask[:, center_y, center_x] = 1.0
                center_reg[:, center_y, center_x] = center - center_int
                center_reg_mask[:, center_y, center_x] = 1.0
        return tuple([nd.array(x) for x in \
            (heatmap, wh_target, wh_mask, center_reg, center_reg_mask)])


def _gaussian_radius(det_size, min_overlap=0.7):
    """Calculate gaussian radius for foreground objects.

    Parameters
    ----------
    det_size : tuple of int
        Object size (h, w).
    min_overlap : float
        Minimal overlap between objects.

    Returns
    -------
    float
        Gaussian radius.

    """
    height, width = det_size

    a1 = 1
    b1 = (height + width)
    c1 = width * height * (1 - min_overlap) / (1 + min_overlap)
    sq1 = np.sqrt(b1 ** 2 - 4 * a1 * c1)
    r1 = (b1 + sq1) / 2

    a2 = 4
    b2 = 2 * (height + width)
    c2 = (1 - min_overlap) * width * height
    sq2 = np.sqrt(b2 ** 2 - 4 * a2 * c2)
    r2 = (b2 + sq2) / 2

    a3 = 4 * min_overlap
    b3 = -2 * min_overlap * (height + width)
    c3 = (min_overlap - 1) * width * height
    sq3 = np.sqrt(b3 ** 2 - 4 * a3 * c3)
    r3 = (b3 + sq3) / 2
    return min(r1, r2, r3)

def _gaussian_2d(shape, sigma=1):
    """Generate 2d gaussian.

    Parameters
    ----------
    shape : tuple of int
        The shape of the gaussian.
    sigma : float
        Sigma for gaussian.

    Returns
    -------
    float
        2D gaussian kernel.

    """
    m, n = [(ss - 1.) / 2. for ss in shape]
    y, x = np.ogrid[-m:m+1, -n:n+1]

    h = np.exp(-(x * x + y * y) / (2 * sigma * sigma))
    h[h < np.finfo(h.dtype).eps * h.max()] = 0
    return h

def _draw_umich_gaussian(heatmap, center, radius, k=1):
    """Draw a 2D gaussian heatmap.

    Parameters
    ----------
    heatmap : numpy.ndarray
        Heatmap to be write inplace.
    center : tuple of int
        Center of object (h, w).
    radius : type
        The radius of gaussian.

    Returns
    -------
    numpy.ndarray
        Drawn gaussian heatmap.

    """
    diameter = 2 * radius + 1
    gaussian = _gaussian_2d((diameter, diameter), sigma=diameter / 6)

    x, y = int(center[0]), int(center[1])

    height, width = heatmap.shape[0:2]

    left, right = min(x, radius), min(width - x, radius + 1)
    top, bottom = min(y, radius), min(height - y, radius + 1)

    masked_heatmap = heatmap[y - top:y + bottom, x - left:x + right]
    masked_gaussian = gaussian[radius - top:radius + bottom, radius - left:radius + right]
    if min(masked_gaussian.shape) > 0 and min(masked_heatmap.shape) > 0:
        np.maximum(masked_heatmap, masked_gaussian * k, out=masked_heatmap)
    return heatmap



import numpy as np
from PIL import Image

import torch

from .utils import images as imageutils


class ClipToTensor(object):
    """Convert a list of m (H x W x C) numpy.ndarrays in the range [0, 255]
    to a torch.FloatTensor of shape (C x m x H x W) in the range [0, 1.0]
    """

    def __init__(self, channel_nb=3, div_255=True, numpy=False):
        self.channel_nb = channel_nb
        self.div_255 = div_255
        self.numpy = numpy

    def __call__(self, clip):
        """
        Args: clip (list of numpy.ndarray): clip (list of images)
        to be converted to tensor.
        """
        # Retrieve shape
        if isinstance(clip[0], np.ndarray):
            h, w, ch = clip[0].shape
            assert ch == self.channel_nb, 'Got {0} instead of 3 channels'.format(
                ch)
        elif isinstance(clip[0], Image.Image):
            w, h = clip[0].size
        else:
            raise TypeError('Expected numpy.ndarray or PIL.Image\
            but got list of {0}'.format(type(clip[0])))

        np_clip = np.zeros([self.channel_nb, len(clip), int(h), int(w)])

        # Convert
        for img_idx, img in enumerate(clip):
            if isinstance(img, np.ndarray):
                pass
            elif isinstance(img, Image.Image):
                img = np.array(img, copy=False)
            else:
                raise TypeError('Expected numpy.ndarray or PIL.Image\
                but got list of {0}'.format(type(clip[0])))
            img = imageutils.convert_img(img)
            np_clip[:, img_idx, :, :] = img
        if self.numpy:
            if self.div_255:
                np_clip = np_clip / 255.0
            return np_clip

        else:
            tensor_clip = torch.from_numpy(np_clip)

            if not isinstance(tensor_clip, torch.FloatTensor):
                tensor_clip = tensor_clip.float()
            if self.div_255:
                tensor_clip = torch.div(tensor_clip, 255)
            return tensor_clip


# Note this norms data to -1/1
class ClipToTensor_K(object):
    """Convert a list of m (H x W x C) numpy.ndarrays in the range [0, 255]
    to a torch.FloatTensor of shape (C x m x H x W) in the range [0, 1.0]
    """

    def __init__(self, channel_nb=3, div_255=True, numpy=False):
        self.channel_nb = channel_nb
        self.div_255 = div_255
        self.numpy = numpy

    def __call__(self, clip):
        """
        Args: clip (list of numpy.ndarray): clip (list of images)
        to be converted to tensor.
        """
        # Retrieve shape
        if isinstance(clip[0], np.ndarray):
            h, w, ch = clip[0].shape
            assert ch == self.channel_nb, 'Got {0} instead of 3 channels'.format(
                ch)
        elif isinstance(clip[0], Image.Image):
            w, h = clip[0].size
        else:
            raise TypeError('Expected numpy.ndarray or PIL.Image\
            but got list of {0}'.format(type(clip[0])))

        np_clip = np.zeros([self.channel_nb, len(clip), int(h), int(w)])

        # Convert
        for img_idx, img in enumerate(clip):
            if isinstance(img, np.ndarray):
                pass
            elif isinstance(img, Image.Image):
                img = np.array(img, copy=False)
            else:
                raise TypeError('Expected numpy.ndarray or PIL.Image\
                but got list of {0}'.format(type(clip[0])))
            img = imageutils.convert_img(img)
            np_clip[:, img_idx, :, :] = img
        if self.numpy:
            if self.div_255:
                np_clip = (np_clip - 127.5) / 127.5
            return np_clip

        else:
            tensor_clip = torch.from_numpy(np_clip)

            if not isinstance(tensor_clip, torch.FloatTensor):
                tensor_clip = tensor_clip.float()
            if self.div_255:
                tensor_clip = torch.div(torch.sub(tensor_clip, 127.5), 127.5)
            return tensor_clip


class ToTensor(object):
    """Converts numpy array to tensor
    """

    def __call__(self, array):
        tensor = torch.from_numpy(array)
        return tensor



"""
MXNet implementation of SMOT: Single-Shot Multi Object Tracking
https://arxiv.org/abs/2010.16031
"""
import os
import numpy as np
import mxnet as mx
from mxnet.gluon.contrib.nn import SyncBatchNorm

from ..model_store import get_model_file
from .ssd_bifpn import get_ssd as get_bifpn_ssd
from .utils import mxnet_frame_preprocessing, timeit_context
from .utils import remap_bboxes as _remap_bboxes


def get_bifpn_net(classes,
                  data_shape,
                  pretrain_weight_file=None,
                  use_keypoints=True,
                  return_features=False,
                  root=os.path.join('~', '.mxnet', 'models'),
                  ctx=None,
                  **kwargs):
    """
    Load bifpn_ssd model with mobilenet backbone
    """

    anchor_sizes = [[.02 * data_shape],
                    [.04 * data_shape, .06 * data_shape],
                    [.08 * data_shape, .12 * data_shape, .16 * data_shape],
                    [.2 * data_shape, .25 * data_shape, .3 * data_shape],
                    [.4 * data_shape, .5 * data_shape],
                    [.6 * data_shape, .75 * data_shape, .9 * data_shape]]

    anchor_ratios = [[1, 1 / 3], [1, 1 / 3, ], [1, 1 / 3, ], [1, 1 / 3, 1.5], [1, 1 / 3, 1.5], [1, 1 / 3, 1.5]]
    pretrained_base = False
    features = ['relu6_fwd', 'relu10_fwd', 'relu22_fwd', 'relu25_fwd']
    ssd_filters = [128, 128]
    fpn_filters = 64
    steps = [4, 8, 16, 32, 64, 128]

    net = get_bifpn_ssd("MobileNet1.0",
                        data_shape,
                        features=features,
                        ssd_filters=ssd_filters,
                        sizes=anchor_sizes,
                        ratios=anchor_ratios,
                        steps=steps,
                        classes=classes,
                        dataset='',
                        pretrained_base=pretrained_base,
                        is_fpn=True,
                        fpn_filters=fpn_filters,
                        is_multitask=False,
                        use_keypoints=use_keypoints,
                        return_features=return_features,
                        **kwargs)

    if pretrain_weight_file is None:
        net.load_parameters(get_model_file('smot_ssd_bifpn_mobilenet', tag=True, root=root), ctx=ctx)
    else:
        net.load_parameters(pretrain_weight_file, ctx=ctx)

    return net


class JointObjectDetector:
    """
    Define the general single shot object detector
    """
    def __init__(self, gpu_id, data_shape,
                 aspect_ratio=16/9., network_param=None):
        self.ctx = mx.gpu(gpu_id)
        self.data_shape = data_shape
        self.net = get_bifpn_net(classes=['face', 'person'],
                                 data_shape=self.data_shape,
                                 pretrain_weight_file=network_param,
                                 ctx=self.ctx,
                                 return_features=False,
                                 use_keypoints=False,
                                 norm_layer=SyncBatchNorm,
                                 norm_kwargs={'num_devices': 1},
                                 )

        self.anchor_tensor = None
        self._anchor_image_shape = (1, 1)
        self._anchor_num = 1

        self.mean_mx = mx.nd.array(np.array([0.485, 0.456, 0.406])).as_in_context(self.ctx)
        self.std_mx = mx.nd.array(np.array([0.229, 0.224, 0.225])).as_in_context(self.ctx)
        self.ratio = aspect_ratio

    def run_detection(self, image, tracking_box_indices, tracking_box_weights, tracking_box_classes):
        """

        Parameters
        ----------
        image: RGB images

        Returns
        -------

        """

        with timeit_context("preprocess"):
            data_tensor, padded_w, padded_h, expand = mxnet_frame_preprocessing(image,
                                                                                self.data_shape,
                                                                                self.ratio,
                                                                                self.mean_mx,
                                                                                self.std_mx,
                                                                                self.ctx)
            mx.nd.waitall()


        with timeit_context("network"):
            real_tracking_indices = tracking_box_indices + tracking_box_classes * self._anchor_num
            ids, scores, detection_bboxes, detection_anchor_indices, tracking_results, anchors = self.net(
                data_tensor.as_in_context(self.ctx), None, real_tracking_indices, tracking_box_weights)

            tracking_bboxes = tracking_results[:, [2, 3, 4, 5, 1]]

            detection_bboxes = _remap_bboxes(detection_bboxes[0, :, :],
                                             padded_w, padded_h, expand,
                                             self.data_shape, self.ratio)
            tracking_bboxes = _remap_bboxes(tracking_bboxes,
                                            padded_w, padded_h, expand,
                                            self.data_shape, self.ratio)
            mx.nd.waitall()

        # set anchors if needed
        if self._anchor_image_shape != (image.shape[:2]):
            self._anchor_image_shape = image.shape[:2]
            # initialize the anchor tensor for assignment
            self.anchor_tensor = anchors[0, :, :]
            half_w = self.anchor_tensor[:, 2] / 2
            half_h = self.anchor_tensor[:, 3] / 2
            center_x = self.anchor_tensor[:, 0].copy()
            center_y = self.anchor_tensor[:, 1].copy()

            # anchors are in the original format of (center_x, center_y, w, h)
            # translate them to (x0, y0, x1, y1)
            self.anchor_tensor[:, 0] = center_x - half_w
            self.anchor_tensor[:, 1] = center_y - half_h

            self.anchor_tensor[:, 2] = center_x + half_w
            self.anchor_tensor[:, 3] = center_y + half_h

            self.anchor_tensor = _remap_bboxes(self.anchor_tensor, padded_w, padded_h, expand,
                                               self.data_shape, self.ratio)
            self._anchor_num = self.anchor_tensor.shape[0]

        return ids[0], scores[0], detection_bboxes, tracking_bboxes, detection_anchor_indices[0].asnumpy()



# pylint: disable=missing-function-docstring, line-too-long, unused-argument
"""
SlowFast Networks for Video Recognition
ICCV 2019, https://arxiv.org/abs/1812.03982
Code adapted from https://github.com/open-mmlab/mmaction and
https://github.com/decisionforce/TPN
"""
from mxnet import init
from mxnet.context import cpu
from mxnet.gluon.block import HybridBlock
from mxnet.gluon import nn
from mxnet.gluon.nn import BatchNorm

from .non_local import build_nonlocal_block


__all__ = ['ResNet_SlowFast', 'i3d_slow_resnet101_f16s4_kinetics700']


def conv3x3x3(in_planes, out_planes, spatial_stride=1, temporal_stride=1, dilation=1):
    "3x3x3 convolution with padding"
    return nn.Conv3D(in_channels=in_planes,
                     channels=out_planes,
                     kernel_size=3,
                     strides=(temporal_stride, spatial_stride, spatial_stride),
                     padding=dilation,
                     dilation=dilation,
                     use_bias=False)


def conv1x3x3(in_planes, out_planes, spatial_stride=1, temporal_stride=1, dilation=1):
    "1x3x3 convolution with padding"
    return nn.Conv3D(in_channels=in_planes,
                     channels=out_planes,
                     kernel_size=(1, 3, 3),
                     strides=(temporal_stride, spatial_stride, spatial_stride),
                     padding=(0, dilation, dilation),
                     dilation=dilation,
                     use_bias=False)


class Bottleneck(HybridBlock):
    """Bottleneck building block for ResNet50, ResNet101 and ResNet152.
    """
    expansion = 4

    def __init__(self,
                 inplanes,
                 planes,
                 spatial_stride=1,
                 temporal_stride=1,
                 dilation=1,
                 downsample=None,
                 if_inflate=True,
                 inflate_style='3x1x1',
                 if_nonlocal=True,
                 nonlocal_cfg=None,
                 norm_layer=BatchNorm,
                 norm_kwargs=None,
                 layer_name='',
                 **kwargs):

        super(Bottleneck, self).__init__()
        assert inflate_style in ['3x1x1', '3x3x3']
        self.inplanes = inplanes
        self.planes = planes
        self.conv1_stride = 1
        self.conv2_stride = spatial_stride
        self.conv1_stride_t = 1
        self.conv2_stride_t = temporal_stride
        self.layer_name = layer_name

        if if_inflate:
            if inflate_style == '3x1x1':
                self.conv1 = nn.Conv3D(in_channels=inplanes,
                                       channels=planes,
                                       kernel_size=(3, 1, 1),
                                       strides=(self.conv1_stride_t, self.conv1_stride, self.conv1_stride),
                                       padding=(1, 0, 0),
                                       use_bias=False)
                self.conv2 = nn.Conv3D(in_channels=planes,
                                       channels=planes,
                                       kernel_size=(1, 3, 3),
                                       strides=(self.conv2_stride_t, self.conv2_stride, self.conv2_stride),
                                       padding=(0, dilation, dilation),
                                       dilation=(1, dilation, dilation),
                                       use_bias=False)
            else:
                self.conv1 = nn.Conv3D(in_channels=inplanes,
                                       channels=planes,
                                       kernel_size=1,
                                       strides=(self.conv1_stride_t, self.conv1_stride, self.conv1_stride),
                                       use_bias=False)
                self.conv2 = nn.Conv3D(in_channels=planes,
                                       channels=planes,
                                       kernel_size=3,
                                       strides=(self.conv2_stride_t, self.conv2_stride, self.conv2_stride),
                                       padding=(1, dilation, dilation),
                                       dilation=(1, dilation, dilation),
                                       use_bias=False)
        else:
            self.conv1 = nn.Conv3D(in_channels=inplanes,
                                   channels=planes,
                                   kernel_size=1,
                                   strides=(1, self.conv1_stride, self.conv1_stride),
                                   use_bias=False)
            self.conv2 = nn.Conv3D(in_channels=planes,
                                   channels=planes,
                                   kernel_size=(1, 3, 3),
                                   strides=(1, self.conv2_stride, self.conv2_stride),
                                   padding=(0, dilation, dilation),
                                   dilation=(1, dilation, dilation),
                                   use_bias=False)

        self.bn1 = norm_layer(in_channels=planes, **({} if norm_kwargs is None else norm_kwargs))
        self.bn2 = norm_layer(in_channels=planes, **({} if norm_kwargs is None else norm_kwargs))
        self.conv3 = nn.Conv3D(in_channels=planes,
                               channels=planes * self.expansion,
                               kernel_size=1,
                               use_bias=False)
        self.bn3 = norm_layer(in_channels=planes * self.expansion, **({} if norm_kwargs is None else norm_kwargs))
        self.relu = nn.Activation('relu')

        self.downsample = downsample
        self.spatial_tride = spatial_stride
        self.temporal_tride = temporal_stride
        self.dilation = dilation

        if if_nonlocal and nonlocal_cfg is not None:
            nonlocal_cfg_ = nonlocal_cfg.copy()
            nonlocal_cfg_['in_channels'] = planes * self.expansion
            self.nonlocal_block = build_nonlocal_block(nonlocal_cfg_)
        else:
            self.nonlocal_block = None

    def hybrid_forward(self, F, x):
        """Hybrid forward of a ResNet bottleneck block"""
        identity = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv3(out)
        out = self.bn3(out)
        if self.downsample is not None:
            identity = self.downsample(x)
        out = F.Activation(out + identity, act_type='relu')

        if self.nonlocal_block is not None:
            out = self.nonlocal_block(out)
        return out


def make_res_layer(block,
                   inplanes,
                   planes,
                   blocks,
                   spatial_stride=1,
                   temporal_stride=1,
                   dilation=1,
                   inflate_freq=1,
                   inflate_style='3x1x1',
                   nonlocal_freq=1,
                   nonlocal_cfg=None,
                   norm_layer=BatchNorm,
                   norm_kwargs=None,
                   layer_name=''):
    inflate_freq = inflate_freq if not isinstance(inflate_freq, int) else (inflate_freq,) * blocks
    nonlocal_freq = nonlocal_freq if not isinstance(nonlocal_freq, int) else (nonlocal_freq,) * blocks
    assert len(inflate_freq) == blocks
    assert len(nonlocal_freq) == blocks

    downsample = None
    if spatial_stride != 1 or inplanes != planes * block.expansion:
        downsample = nn.HybridSequential(prefix=layer_name+'downsample_')
        with downsample.name_scope():
            downsample.add(nn.Conv3D(in_channels=inplanes,
                                     channels=planes * block.expansion,
                                     kernel_size=1,
                                     strides=(temporal_stride, spatial_stride, spatial_stride),
                                     use_bias=False))
            downsample.add(norm_layer(in_channels=planes * block.expansion, **({} if norm_kwargs is None else norm_kwargs)))


    layers = nn.HybridSequential(prefix=layer_name)
    cnt = 0
    with layers.name_scope():
        layers.add(block(inplanes=inplanes,
                         planes=planes,
                         spatial_stride=spatial_stride,
                         temporal_stride=temporal_stride,
                         dilation=dilation,
                         downsample=downsample,
                         if_inflate=(inflate_freq[0] == 1),
                         inflate_style=inflate_style,
                         if_nonlocal=(nonlocal_freq[0] == 1),
                         nonlocal_cfg=nonlocal_cfg,
                         layer_name='%d_' % cnt))

        cnt += 1
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.add(block(inplanes=inplanes,
                             planes=planes,
                             spatial_stride=1,
                             temporal_stride=1,
                             dilation=dilation,
                             if_inflate=(inflate_freq[i] == 1),
                             inflate_style=inflate_style,
                             if_nonlocal=(nonlocal_freq[i] == 1),
                             nonlocal_cfg=nonlocal_cfg,
                             layer_name='%d_' % cnt))
            cnt += 1
    return layers


class ResNet_SlowFast(HybridBlock):
    """ResNe(x)t_SlowFast backbone.
    Args:
        depth (int): Depth of resnet, from {50, 101, 152}.
        num_stages (int): Resnet stages, normally 4.
        strides (Sequence[int]): Strides of the first block of each stage.
        dilations (Sequence[int]): Dilation of each stage.
        out_indices (Sequence[int]): Output from which stages.
        frozen_stages (int): Stages to be frozen (all param fixed). -1 means
            not freezing any parameters.
        bn_eval (bool): Whether to set BN layers to eval mode, namely, freeze
            running stats (mean and var).
        bn_frozen (bool): Whether to freeze weight and bias of BN layers.
    """

    arch_settings = {
        50: (Bottleneck, (3, 4, 6, 3)),
        101: (Bottleneck, (3, 4, 23, 3)),
        152: (Bottleneck, (3, 8, 36, 3))
    }

    def __init__(self,
                 num_classes,
                 depth,
                 pretrained=None,
                 pretrained_base=True,
                 feat_ext=False,
                 num_segments=1,
                 num_crop=1,
                 num_stages=4,
                 spatial_strides=(1, 2, 2, 2),
                 temporal_strides=(1, 1, 1, 1),
                 dilations=(1, 1, 1, 1),
                 out_indices=(0, 1, 2, 3),
                 conv1_kernel_t=1,
                 conv1_stride_t=1,
                 pool1_kernel_t=1,
                 pool1_stride_t=1,
                 frozen_stages=-1,
                 inflate_freq=(0, 0, 1, 1),
                 inflate_stride=(1, 1, 1, 1),
                 inflate_style='3x1x1',
                 nonlocal_stages=(-1,),
                 nonlocal_freq=(0, 0, 0, 0),
                 nonlocal_cfg=None,
                 bn_eval=False,
                 bn_frozen=False,
                 partial_bn=False,
                 dropout_ratio=0.5,
                 init_std=0.01,
                 norm_layer=BatchNorm,
                 norm_kwargs=None,
                 ctx=None,
                 **kwargs):
        super(ResNet_SlowFast, self).__init__()

        if depth not in self.arch_settings:
            raise KeyError('invalid depth {} for resnet'.format(depth))

        self.num_classes = num_classes
        self.depth = depth
        self.pretrained = pretrained
        self.pretrained_base = pretrained_base
        self.feat_ext = feat_ext
        self.num_segments = num_segments
        self.num_crop = num_crop
        self.num_stages = num_stages
        assert 1 <= num_stages <= 4
        self.spatial_strides = spatial_strides
        self.temporal_strides = temporal_strides
        self.dilations = dilations
        assert len(spatial_strides) == len(temporal_strides) == len(dilations) == num_stages
        self.out_indices = out_indices
        assert max(out_indices) < num_stages
        self.frozen_stages = frozen_stages
        self.inflate_freqs = inflate_freq if not isinstance(inflate_freq, int) else (inflate_freq,) * num_stages
        self.inflate_style = inflate_style
        self.nonlocal_stages = nonlocal_stages
        self.nonlocal_freqs = nonlocal_freq if not isinstance(nonlocal_freq, int) else (nonlocal_freq,) * num_stages
        self.nonlocal_cfg = nonlocal_cfg
        self.bn_eval = bn_eval
        self.bn_frozen = bn_frozen
        self.partial_bn = partial_bn
        self.feat_ext = feat_ext

        self.dropout_ratio = dropout_ratio
        self.init_std = init_std

        self.block, stage_blocks = self.arch_settings[depth]
        self.stage_blocks = stage_blocks[:num_stages]
        self.inplanes = 64

        self.first_stage = nn.HybridSequential(prefix='')
        self.first_stage.add(nn.Conv3D(in_channels=3, channels=64, kernel_size=(conv1_kernel_t, 7, 7),
                                       strides=(conv1_stride_t, 2, 2), padding=((conv1_kernel_t - 1)//2, 3, 3), use_bias=False))
        self.first_stage.add(norm_layer(in_channels=64, **({} if norm_kwargs is None else norm_kwargs)))
        self.first_stage.add(nn.Activation('relu'))
        self.first_stage.add(nn.MaxPool3D(pool_size=(pool1_kernel_t, 3, 3), strides=(pool1_stride_t, 2, 2), padding=(pool1_kernel_t//2, 1, 1)))

        self.res_layers = nn.HybridSequential(prefix='')
        for i, num_blocks in enumerate(self.stage_blocks):
            spatial_stride = spatial_strides[i]
            temporal_stride = temporal_strides[i]
            dilation = dilations[i]
            planes = 64 * 2**i
            layer_name = 'layer{}_'.format(i + 1)

            res_layer = make_res_layer(self.block,
                                       self.inplanes,
                                       planes,
                                       num_blocks,
                                       spatial_stride=spatial_stride,
                                       temporal_stride=temporal_stride,
                                       dilation=dilation,
                                       inflate_freq=self.inflate_freqs[i],
                                       inflate_style=self.inflate_style,
                                       nonlocal_freq=self.nonlocal_freqs[i],
                                       nonlocal_cfg=self.nonlocal_cfg if i in self.nonlocal_stages else None,
                                       norm_layer=norm_layer,
                                       norm_kwargs=norm_kwargs,
                                       layer_name=layer_name)
            self.inplanes = planes * self.block.expansion
            self.res_layers.add(res_layer)

        self.feat_dim = self.block.expansion * 64 * 2 ** (len(self.stage_blocks) - 1)

        self.st_avg = nn.GlobalAvgPool3D()

        self.head = nn.HybridSequential(prefix='')
        self.head.add(nn.Dropout(rate=self.dropout_ratio))
        self.fc = nn.Dense(in_units=self.feat_dim, units=num_classes, weight_initializer=init.Normal(sigma=self.init_std))
        self.head.add(self.fc)

        self.init_weights(ctx)

    def init_weights(self, ctx):
        """Initial I3D_slow network."""
        self.first_stage.initialize(ctx=ctx)
        self.res_layers.initialize(ctx=ctx)
        self.head.initialize(ctx=ctx)

    def hybrid_forward(self, F, x):
        """Hybrid forward of I3D_slow network"""
        x = self.first_stage(x)
        outs = []
        for i, res_layer in enumerate(self.res_layers):
            x = res_layer(x)
            if i in self.out_indices:
                outs.append(x)
        feat = outs[0]

        # spatial temporal average
        pooled_feat = self.st_avg(feat)
        x = F.squeeze(pooled_feat, axis=(2, 3, 4))

        # segmental consensus
        x = F.reshape(x, shape=(-1, self.num_segments * self.num_crop, self.feat_dim))
        x = F.mean(x, axis=1)

        if self.feat_ext:
            return x

        out = self.head(x)
        return out


def i3d_slow_resnet101_f16s4_kinetics700(nclass=700, pretrained=False, pretrained_base=True, ctx=cpu(),
                                         root='~/.mxnet/models', num_segments=1, num_crop=1,
                                         partial_bn=False, bn_frozen=False, feat_ext=False, **kwargs):
    model = ResNet_SlowFast(num_classes=nclass,
                            depth=101,
                            pretrained=pretrained,
                            pretrained_base=pretrained_base,
                            feat_ext=feat_ext,
                            num_segments=num_segments,
                            num_crop=num_crop,
                            out_indices=[3],
                            bn_eval=False,
                            partial_bn=partial_bn,
                            bn_frozen=bn_frozen,
                            ctx=ctx,
                            **kwargs)

    if pretrained:
        from ..model_store import get_model_file
        model.load_parameters(get_model_file('i3d_slow_resnet101_f16s4_kinetics700',
                                             tag=pretrained, root=root), ctx=ctx)
        from ...data import Kinetics700Attr
        attrib = Kinetics700Attr()
        model.classes = attrib.classes
    model.collect_params().reset_ctx(ctx)

    return model



# pylint: disable=missing-function-docstring, missing-class-docstring
"""
A Closer Look at Spatiotemporal Convolutions for Action Recognition
CVPR 2018, https://arxiv.org/abs/1711.11248
Large-scale weakly-supervised pre-training for video action recognition
CVPR 2019, https://arxiv.org/abs/1905.00561
"""
import torch
import torch.nn as nn


__all__ = ['ResNet_R2plus1Dv2', 'r2plus1d_v2_resnet152_kinetics400']


eps = 1e-3
bn_mmt = 0.1


class Affine(nn.Module):
    def __init__(self, feature_in):
        super(Affine, self).__init__()
        self.weight = nn.Parameter(torch.randn(feature_in, 1, 1, 1))
        self.bias = nn.Parameter(torch.randn(feature_in, 1, 1, 1))
        self.weight.requires_grad = True
        self.bias.requires_grad = True

    def forward(self, x):
        x = x * self.weight + self.bias
        return x


class Bottleneck_R2plus1Dv2(nn.Module):
    def __init__(self, in_planes, planes, middle_planes, stride=1, temporal_stride=1,
                 down_sample=None, expansion=2, temporal_kernel=3, use_affine=True):
        super(Bottleneck_R2plus1Dv2, self).__init__()
        self.expansion = expansion
        self.conv1 = nn.Conv3d(in_planes, planes, kernel_size=(1, 1, 1),
                               bias=False, stride=(1, 1, 1))

        if use_affine:
            self.bn1 = Affine(planes)
        else:
            self.bn1 = nn.BatchNorm3d(planes, track_running_stats=True,
                                      eps=eps, momentum=bn_mmt)

        self.conv2_middle = nn.Conv3d(
            planes,
            middle_planes,
            kernel_size=(1, 3, 3),
            stride=(1, stride, stride),
            padding=(0, 1, 1),
            groups=1,
            bias=False)
        if use_affine:
            self.bn2_middle = Affine(middle_planes)
        else:
            self.bn2_middle = nn.BatchNorm3d(middle_planes, track_running_stats=True,
                                             eps=eps, momentum=bn_mmt)

        self.conv2 = nn.Conv3d(middle_planes, planes, kernel_size=(temporal_kernel, 1, 1),
                               bias=False, stride=(temporal_stride, 1, 1),
                               padding=((temporal_kernel - 1) // 2, 0, 0))

        if use_affine:
            self.bn2 = Affine(planes)
        else:
            self.bn2 = nn.BatchNorm3d(planes, track_running_stats=True, eps=eps, momentum=bn_mmt)

        self.conv3 = nn.Conv3d(
            planes, planes * self.expansion, kernel_size=1, bias=False)

        if use_affine:
            self.bn3 = Affine(planes * self.expansion)
        else:
            self.bn3 = nn.BatchNorm3d(planes * self.expansion, track_running_stats=True,
                                      eps=eps, momentum=bn_mmt)

        self.relu = nn.ReLU(inplace=True)
        self.down_sample = down_sample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2_middle(out)
        out = self.bn2_middle(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.down_sample is not None:
            residual = self.down_sample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet_R2plus1Dv2(nn.Module):
    def __init__(self,
                 block,
                 block_nums,
                 num_classes=400,
                 feat_ext=False,
                 use_affine=True):

        self.use_affine = use_affine
        self.in_planes = 64
        self.num_classes = num_classes
        self.feat_ext = feat_ext

        super(ResNet_R2plus1Dv2, self).__init__()
        self.conv1_middle = nn.Conv3d(
            3,
            45,
            kernel_size=(1, 7, 7),
            stride=(1, 2, 2),
            padding=(0, 3, 3),
            bias=False)
        if use_affine:
            self.bn1_middle = Affine(45)
        else:
            self.bn1_middle = nn.BatchNorm3d(45, track_running_stats=True,
                                             eps=eps, momentum=bn_mmt)

        self.conv1 = nn.Conv3d(
            45,
            64,
            kernel_size=(3, 1, 1),
            stride=(1, 1, 1),
            padding=(1, 0, 0),
            bias=False)
        if use_affine:
            self.bn1 = Affine(64)
        else:
            self.bn1 = nn.BatchNorm3d(64, track_running_stats=True, eps=eps, momentum=bn_mmt)

        self.relu = nn.ReLU(inplace=True)

        self.layer1 = self._make_layer(block,
                                       in_planes=64,
                                       planes=64,
                                       middle_planes=144,
                                       blocks=block_nums[0],
                                       stride=1,
                                       expansion=4)

        self.layer2 = self._make_layer(block,
                                       in_planes=256,
                                       planes=128,
                                       middle_planes=288,
                                       blocks=block_nums[1],
                                       stride=2,
                                       temporal_stride=2,
                                       expansion=4)

        self.layer3 = self._make_layer(block,
                                       in_planes=512,
                                       planes=256,
                                       middle_planes=576,
                                       blocks=block_nums[2],
                                       stride=2,
                                       temporal_stride=2,
                                       expansion=4)

        self.layer4 = self._make_layer(block,
                                       in_planes=1024,
                                       planes=512,
                                       middle_planes=1152,
                                       blocks=block_nums[3],
                                       stride=2,
                                       temporal_stride=2,
                                       expansion=4)

        self.avgpool = nn.AdaptiveAvgPool3d(output_size=(1, 1, 1))

        self.out_fc = nn.Linear(in_features=2048, out_features=num_classes, bias=True)

    def _make_layer(self,
                    block,
                    in_planes,
                    planes,
                    middle_planes,
                    blocks,
                    stride=1,
                    temporal_stride=1,
                    expansion=4):

        if self.use_affine:
            down_bn = Affine(planes * expansion)
        else:
            down_bn = nn.BatchNorm3d(planes * expansion, track_running_stats=True,
                                     eps=eps, momentum=bn_mmt)
        down_sample = nn.Sequential(
            nn.Conv3d(
                in_planes,
                planes * expansion,
                kernel_size=1,
                stride=(temporal_stride, stride, stride),
                bias=False), down_bn)
        layers = []
        layers.append(
            block(in_planes, planes, middle_planes, stride, temporal_stride,
                  down_sample, expansion, temporal_kernel=3, use_affine=self.use_affine))
        for _ in range(1, blocks):
            layers.append(block(planes * expansion, planes, middle_planes, expansion=expansion,
                                temporal_kernel=3, use_affine=self.use_affine))

        return nn.Sequential(*layers)

    def forward(self, x):
        bs, _, _, _, _ = x.size()
        x = self.conv1_middle(x)
        x = self.bn1_middle(x)
        x = self.relu(x)

        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(bs, -1)

        if self.feat_ext:
            return x

        logits = self.out_fc(x)
        return logits


def r2plus1d_v2_resnet152_kinetics400(cfg):
    model = ResNet_R2plus1Dv2(Bottleneck_R2plus1Dv2,
                              num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                              block_nums=[3, 8, 36, 3],
                              feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                              use_affine=cfg.CONFIG.MODEL.USE_AFFINE)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('r2plus1d_v2_resnet152_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model



"""Decoder module of Monodepth2
Code partially borrowed from
https://github.com/nianticlabs/monodepth2/blob/master/networks/pose_decoder.py
"""
# Copyright Niantic 2019. Patent Pending. All rights reserved.
#
# This software is licensed under the terms of the Monodepth2 licence
# which allows for non-commercial use only, the full terms of which are made
# available in the LICENSE file.

from __future__ import absolute_import, division, print_function

from collections import OrderedDict
import mxnet as mx
import mxnet.gluon.nn as nn


class PoseDecoder(nn.HybridBlock):
    r"""Decoder of Monodepth2 PoseNet

    Parameters
    ----------
    num_ch_enc : list
        The channels number of encoder.
    num_input_features: int
        The number of input sequences. 1 for depth encoder, larger than 1 for pose encoder.
        (Default: 2)
    num_frames_to_predict_for: int
        The number of output pose between frames; If None, it equals num_input_features - 1.
        (Default: 2)
    stride: int
        The stride number for Conv in pose decoder. (Default: 1)

    """

    def __init__(self, num_ch_enc, num_input_features, num_frames_to_predict_for=2, stride=1):
        super(PoseDecoder, self).__init__()

        self.num_ch_enc = num_ch_enc
        self.num_input_features = num_input_features

        if num_frames_to_predict_for is None:
            num_frames_to_predict_for = num_input_features - 1
        self.num_frames_to_predict_for = num_frames_to_predict_for

        self.convs = OrderedDict()
        self.convs[("squeeze")] = nn.Conv2D(
            in_channels=self.num_ch_enc[-1], channels=256, kernel_size=1)
        self.convs[("pose", 0)] = nn.Conv2D(
            in_channels=num_input_features * 256, channels=256,
            kernel_size=3, strides=stride, padding=1)
        self.convs[("pose", 1)] = nn.Conv2D(
            in_channels=256, channels=256, kernel_size=3, strides=stride, padding=1)
        self.convs[("pose", 2)] = nn.Conv2D(
            in_channels=256, channels=6 * num_frames_to_predict_for, kernel_size=1)

        # register blocks
        for k in self.convs:
            self.register_child(self.convs[k])
        self.net = nn.HybridSequential()
        self.net.add(*list(self.convs.values()))

    def hybrid_forward(self, F, input_features):
        # pylint: disable=unused-argument, missing-function-docstring
        last_features = [f[-1] for f in input_features]

        cat_features = [F.relu(self.convs["squeeze"](f)) for f in last_features]
        cat_features = F.concat(*cat_features, dim=1)

        out = cat_features
        for i in range(3):
            out = self.convs[("pose", i)](out)
            if i != 2:
                out = F.relu(out)

        out = out.mean(3).mean(2)

        out = 0.01 * out.reshape(shape=(-1, self.num_frames_to_predict_for, 1, 6))

        axisangle = F.slice(out, begin=(0, 0, 0, 0), end=(None, 2, 1, 3))
        translation = F.slice(out, begin=(0, 0, 0, 3), end=(None, 2, 1, 6))

        return axisangle, translation

    def predict(self, input_features):
        # pylint: disable=unused-argument, missing-function-docstring
        last_features = [f[-1] for f in input_features]

        cat_features = [mx.nd.relu()(self.convs["squeeze"](f)) for f in last_features]
        cat_features = mx.nd.concat(*cat_features, dim=1)

        out = cat_features
        for i in range(3):
            out = self.convs[("pose", i)](out)
            if i != 2:
                out = mx.nd.relu(out)

        out = out.mean(3).mean(2)

        out = 0.01 * out.reshape(-1, self.num_frames_to_predict_for, 1, 6)

        axisangle = out[..., :3]
        translation = out[..., 3:]

        return axisangle, translation



"""Customized dataloader for general video classification tasks."""
import os
import warnings
import numpy as np
from decord import VideoReader, cpu

import torch
from torch.utils.data import Dataset

from ..transforms.videotransforms import video_transforms, volume_transforms
from .multigrid_helper import multiGridSampler, MultiGridBatchSampler


__all__ = ['VideoClsDataset', 'build_dataloader', 'build_dataloader_test']


class VideoClsDataset(Dataset):
    """Load your own video classification dataset."""

    def __init__(self, anno_path, data_path, mode='train', clip_len=8,
                 frame_sample_rate=2, crop_size=224, short_side_size=256,
                 new_height=256, new_width=340, keep_aspect_ratio=False,
                 num_segment=1, num_crop=1, test_num_segment=10, test_num_crop=3,
                 use_multigrid=False):
        self.anno_path = anno_path
        self.data_path = data_path
        self.mode = mode
        self.clip_len = clip_len
        self.frame_sample_rate = frame_sample_rate
        self.crop_size = crop_size
        self.short_side_size = short_side_size
        self.new_height = new_height
        self.new_width = new_width
        self.keep_aspect_ratio = keep_aspect_ratio
        self.num_segment = num_segment
        self.test_num_segment = test_num_segment
        self.num_crop = num_crop
        self.test_num_crop = test_num_crop
        self.use_multigrid = use_multigrid and (mode == 'train')

        import pandas as pd
        cleaned = pd.read_csv(self.anno_path, header=None, delimiter=' ')
        self.dataset_samples = list(cleaned.values[:, 0])
        self.label_array = list(cleaned.values[:, 2])

        if (mode == 'train'):
            if self.use_multigrid:
                self.MG_sampler = multiGridSampler()
                self.data_transform = []
                for alpha in range(self.MG_sampler.mod_long):
                    tmp = []
                    for beta in range(self.MG_sampler.mod_short):
                        info = self.MG_sampler.get_resize(alpha, beta)
                        scale_s = info[1]
                        tmp.append(video_transforms.Compose([
                            video_transforms.Resize(int(self.short_side_size / scale_s),
                                                    interpolation='bilinear'),
                            # TODO: multiscale corner cropping
                            video_transforms.RandomResize(ratio=(1, 1.25),
                                                          interpolation='bilinear'),
                            video_transforms.RandomCrop(size=(int(self.crop_size / scale_s),
                                                              int(self.crop_size / scale_s)))]))
                    self.data_transform.append(tmp)
            else:
                self.data_transform = video_transforms.Compose([
                    video_transforms.Resize(int(self.short_side_size),
                                            interpolation='bilinear'),
                    video_transforms.RandomResize(ratio=(1, 1.25),
                                                  interpolation='bilinear'),
                    video_transforms.RandomCrop(size=(int(self.crop_size),
                                                      int(self.crop_size)))])

            self.data_transform_after = video_transforms.Compose([
                video_transforms.RandomHorizontalFlip(),
                volume_transforms.ClipToTensor(),
                video_transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                           std=[0.229, 0.224, 0.225])
            ])
        elif (mode == 'validation'):
            self.data_transform = video_transforms.Compose([
                video_transforms.Resize(self.short_side_size, interpolation='bilinear'),
                video_transforms.CenterCrop(size=(self.crop_size, self.crop_size)),
                volume_transforms.ClipToTensor(),
                video_transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                           std=[0.229, 0.224, 0.225])
            ])
        elif mode == 'test':
            self.data_resize = video_transforms.Compose([
                video_transforms.Resize(size=(short_side_size), interpolation='bilinear')
            ])
            self.data_transform = video_transforms.Compose([
                volume_transforms.ClipToTensor(),
                video_transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                           std=[0.229, 0.224, 0.225])
            ])
            self.test_seg = []
            self.test_dataset = []
            self.test_label_array = []
            for ck in range(self.test_num_segment):
                for cp in range(self.test_num_crop):
                    for idx in range(len(self.label_array)):
                        sample_label = self.label_array[idx]
                        self.test_label_array.append(sample_label)
                        self.test_dataset.append(self.dataset_samples[idx])
                        self.test_seg.append((ck, cp))

    def __getitem__(self, index):
        if self.mode == 'train':
            if self.use_multigrid is True:
                index, alpha, beta = index
                info = self.MG_sampler.get_resize(alpha, beta)
                scale_t = info[0]
                data_transform_func = self.data_transform[alpha][beta]
            else:
                scale_t = 1
                data_transform_func = self.data_transform

            sample = self.dataset_samples[index]
            buffer = self.loadvideo_decord(sample, sample_rate_scale=scale_t)
            if len(buffer) == 0:
                while len(buffer) == 0:
                    warnings.warn("video {} not correctly loaded during training".format(sample))
                    index = np.random.randint(self.__len__())
                    sample = self.dataset_samples[index]
                    buffer = self.loadvideo_decord(sample, sample_rate_scale=scale_t)

            buffer = data_transform_func(buffer)
            buffer = self.data_transform_after(buffer)
            return buffer, self.label_array[index], sample.split("/")[-1].split(".")[0]

        elif self.mode == 'validation':
            sample = self.dataset_samples[index]
            buffer = self.loadvideo_decord(sample)
            if len(buffer) == 0:
                while len(buffer) == 0:
                    warnings.warn("video {} not correctly loaded during validation".format(sample))
                    index = np.random.randint(self.__len__())
                    sample = self.dataset_samples[index]
                    buffer = self.loadvideo_decord(sample)
            buffer = self.data_transform(buffer)
            return buffer, self.label_array[index], sample.split("/")[1].split(".")[0]

        elif self.mode == 'test':
            sample = self.test_dataset[index]
            chunk_nb, split_nb = self.test_seg[index]
            buffer = self.loadvideo_decord(sample)

            while len(buffer) == 0:
                warnings.warn("video {}, temporal {}, spatial {} not found during testing".format(\
                    str(self.test_dataset[index]), chunk_nb, split_nb))
                index = np.random.randint(self.__len__())
                sample = self.test_dataset[index]
                chunk_nb, split_nb = self.test_seg[index]
                buffer = self.loadvideo_decord(sample)

            buffer = self.data_resize(buffer)
            if isinstance(buffer, list):
                buffer = np.stack(buffer, 0)

            spatial_step = 1.0 * (max(buffer.shape[1], buffer.shape[2]) - self.short_side_size) \
                                 / (self.test_num_crop - 1)
            temporal_step = max(1.0 * (buffer.shape[0] - self.clip_len) \
                                / (self.test_num_segment - 1), 0)
            temporal_start = int(chunk_nb * temporal_step)
            spatial_start = int(split_nb * spatial_step)
            if buffer.shape[1] >= buffer.shape[2]:
                buffer = buffer[temporal_start:temporal_start + self.clip_len, \
                       spatial_start:spatial_start + self.short_side_size, :, :]
            else:
                buffer = buffer[temporal_start:temporal_start + self.clip_len, \
                       :, spatial_start:spatial_start + self.short_side_size, :]

            buffer = self.data_transform(buffer)
            return buffer, self.test_label_array[index], sample.split("/")[-1].split(".")[0], \
                   chunk_nb, split_nb
        else:
            raise NameError('mode {} unkown'.format(self.mode))

    def loadvideo_decord(self, sample, sample_rate_scale=1):
        """Load video content using Decord"""
        # pylint: disable=line-too-long, bare-except, unnecessary-comprehension
        fname = self.data_path + sample

        if not (os.path.exists(fname)):
            return []

        # avoid hanging issue
        if os.path.getsize(fname) < 1 * 1024:
            print('SKIP: ', fname, " - ", os.path.getsize(fname))
            return []
        try:
            if self.keep_aspect_ratio:
                vr = VideoReader(fname, num_threads=1, ctx=cpu(0))
            else:
                vr = VideoReader(fname, width=self.new_width, height=self.new_height,
                                 num_threads=1, ctx=cpu(0))
        except:
            print("video cannot be loaded by decord: ", fname)
            return []

        if self.mode == 'test':
            all_index = [x for x in range(0, len(vr), self.frame_sample_rate)]
            while len(all_index) < self.clip_len:
                all_index.append(all_index[-1])
            vr.seek(0)
            buffer = vr.get_batch(all_index).asnumpy()
            return buffer

        # handle temporal segments
        converted_len = int(self.clip_len * self.frame_sample_rate)
        seg_len = len(vr) // self.num_segment

        all_index = []
        for i in range(self.num_segment):
            if seg_len <= converted_len:
                index = np.linspace(0, seg_len, num=seg_len // self.frame_sample_rate)
                index = np.concatenate((index, np.ones(self.clip_len - seg_len // self.frame_sample_rate) * seg_len))
                index = np.clip(index, 0, seg_len - 1).astype(np.int64)
            else:
                end_idx = np.random.randint(converted_len, seg_len)
                str_idx = end_idx - converted_len
                index = np.linspace(str_idx, end_idx, num=self.clip_len)
                index = np.clip(index, str_idx, end_idx - 1).astype(np.int64)
            index = index + i*seg_len
            all_index.extend(list(index))

        all_index = all_index[::int(sample_rate_scale)]
        vr.seek(0)
        buffer = vr.get_batch(all_index).asnumpy()
        return buffer

    def __len__(self):
        if self.mode != 'test':
            return len(self.dataset_samples)
        else:
            return len(self.test_dataset)


def build_dataloader(cfg):
    """Build dataloader for training/validation"""
    train_dataset = VideoClsDataset(anno_path=cfg.CONFIG.DATA.TRAIN_ANNO_PATH,
                                    data_path=cfg.CONFIG.DATA.TRAIN_DATA_PATH,
                                    mode='train',
                                    use_multigrid=cfg.CONFIG.DATA.MULTIGRID,
                                    clip_len=cfg.CONFIG.DATA.CLIP_LEN,
                                    frame_sample_rate=cfg.CONFIG.DATA.FRAME_RATE,
                                    num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                                    num_crop=cfg.CONFIG.DATA.NUM_CROP,
                                    keep_aspect_ratio=cfg.CONFIG.DATA.KEEP_ASPECT_RATIO,
                                    crop_size=cfg.CONFIG.DATA.CROP_SIZE,
                                    short_side_size=cfg.CONFIG.DATA.SHORT_SIDE_SIZE,
                                    new_height=cfg.CONFIG.DATA.NEW_HEIGHT,
                                    new_width=cfg.CONFIG.DATA.NEW_WIDTH)
    val_dataset = VideoClsDataset(anno_path=cfg.CONFIG.DATA.VAL_ANNO_PATH,
                                  data_path=cfg.CONFIG.DATA.VAL_DATA_PATH,
                                  mode='validation',
                                  use_multigrid=cfg.CONFIG.DATA.MULTIGRID,
                                  clip_len=cfg.CONFIG.DATA.CLIP_LEN,
                                  frame_sample_rate=cfg.CONFIG.DATA.FRAME_RATE,
                                  num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                                  num_crop=cfg.CONFIG.DATA.NUM_CROP,
                                  keep_aspect_ratio=cfg.CONFIG.DATA.KEEP_ASPECT_RATIO,
                                  crop_size=cfg.CONFIG.DATA.CROP_SIZE,
                                  short_side_size=cfg.CONFIG.DATA.SHORT_SIDE_SIZE,
                                  new_height=cfg.CONFIG.DATA.NEW_HEIGHT,
                                  new_width=cfg.CONFIG.DATA.NEW_WIDTH)

    if cfg.DDP_CONFIG.DISTRIBUTED:
        train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
        val_sampler = torch.utils.data.distributed.DistributedSampler(val_dataset)
    else:
        train_sampler = None
        val_sampler = None

    mg_sampler = None
    if cfg.CONFIG.DATA.MULTIGRID:
        mg_sampler = MultiGridBatchSampler(train_sampler, batch_size=cfg.CONFIG.TRAIN.BATCH_SIZE,
                                           drop_last=True)
        train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=1, shuffle=False,
                                                   num_workers=9, pin_memory=True,
                                                   batch_sampler=mg_sampler)
    else:
        train_loader = torch.utils.data.DataLoader(
            train_dataset, batch_size=cfg.CONFIG.TRAIN.BATCH_SIZE, shuffle=(train_sampler is None),
            num_workers=9, sampler=train_sampler, pin_memory=True)

    val_loader = torch.utils.data.DataLoader(
        val_dataset, batch_size=cfg.CONFIG.VAL.BATCH_SIZE, shuffle=(val_sampler is None),
        num_workers=9, sampler=val_sampler, pin_memory=True)

    return train_loader, val_loader, train_sampler, val_sampler, mg_sampler

def build_dataloader_test(cfg):
    """Build dataloader for testing"""
    test_dataset = VideoClsDataset(anno_path=cfg.CONFIG.DATA.VAL_ANNO_PATH,
                                   data_path=cfg.CONFIG.DATA.VAL_DATA_PATH,
                                   mode='test',
                                   clip_len=cfg.CONFIG.DATA.CLIP_LEN,
                                   frame_sample_rate=cfg.CONFIG.DATA.FRAME_RATE,
                                   test_num_segment=cfg.CONFIG.DATA.TEST_NUM_SEGMENT,
                                   test_num_crop=cfg.CONFIG.DATA.TEST_NUM_CROP,
                                   keep_aspect_ratio=cfg.CONFIG.DATA.KEEP_ASPECT_RATIO,
                                   crop_size=cfg.CONFIG.DATA.CROP_SIZE,
                                   short_side_size=cfg.CONFIG.DATA.SHORT_SIDE_SIZE,
                                   new_height=cfg.CONFIG.DATA.NEW_HEIGHT,
                                   new_width=cfg.CONFIG.DATA.NEW_WIDTH)

    if cfg.DDP_CONFIG.DISTRIBUTED:
        test_sampler = torch.utils.data.distributed.DistributedSampler(test_dataset)
    else:
        test_sampler = None
    test_loader = torch.utils.data.DataLoader(
        test_dataset, batch_size=cfg.CONFIG.VAL.BATCH_SIZE, shuffle=(test_sampler is None),
        num_workers=9, sampler=test_sampler, pin_memory=True)

    return test_loader



# pylint: disable=line-too-long,too-many-lines,missing-docstring
import random

import numpy as np
from mxnet import gluon

__all__ = ['SplitSampler', 'ShuffleSplitSampler', 'SplitSortedBucketSampler']


class SplitSampler(gluon.data.sampler.Sampler):
    """ Split the dataset into `num_parts` parts and sample from the part with index `part_index`

    Parameters
    ----------
    length: int
      Number of examples in the dataset
    num_parts: int
      Partition the data into multiple parts
    part_index: int
      The index of the part to read from
    """

    def __init__(self, length, num_parts=1, part_index=0):
        # Compute the length of each partition
        self.part_len = length // num_parts
        # Compute the start index for this partition
        self.start = self.part_len * part_index
        # Compute the end index for this partition
        self.end = self.start + self.part_len

    def __iter__(self):
        # Extract examples between `start` and `end`, shuffle and return them.
        indices = list(range(self.start, self.end))
        random.shuffle(indices)
        return iter(indices)

    def __len__(self):
        return self.part_len


class ShuffleSplitSampler(gluon.data.sampler.Sampler):
    """Split the dataset into `num_parts` parts and randomly sample from the part
    with index `part_index`.
    The data is randomly shuffled at each iteration within each partition.

    Parameters
    ----------
    length: int
      Number of examples in the dataset
    num_parts: int
      Number of partitions which the data is split into
    part_index: int
      The index of the part to read from
    """

    def __init__(self, length, num_parts=1, part_index=0, seed=0):
        if length % num_parts != 0:
            print('Length ({}) must be a multiple of the number of partitions ({}).'.format(length,
                                                                                            num_parts))
        self._seed = seed
        self._state = np.random.RandomState(seed)
        self._indices = list(range(length))
        # Compute the length of each partition
        part_len = length // num_parts
        # Compute the start index for this partition
        self._start = part_len * part_index
        # Compute the end index for this partition
        self._end = self._start + part_len
        if part_index == num_parts - 1:
            self._end = length

    def __iter__(self):
        self._state.shuffle(self._indices)
        # Extract examples between `start` and `end`, shuffle and return them.
        indices = list(self._indices[self._start:self._end])
        return iter(indices)

    def __len__(self):
        return self._end - self._start


class SplitSortedBucketSampler(gluon.data.sampler.Sampler):
    r"""Batches are sampled from sorted buckets of data.
    First, partition data in buckets of size `batch_size * mult`.
    Each bucket contains `batch_size * mult` elements. The samples inside each bucket are sorted
    based on sort_key and then batched.
    Parameters
    ----------
    sort_keys : list-like object
        The keys to sort the samples.
    batch_size : int
        Batch size of the sampler.
    mult : int or float, default 100
        The multiplier to determine the bucket size. Each bucket will have size `mult * batch_size`.
    num_parts: int, default 1
      Number of partitions which the data is split into
    part_index: int, default 0
      The index of the part to read from
    shuffle : bool, default False
        Whether to shuffle the data.
    Examples
    --------
    >>> lengths = [np.random.randint(1, 1000) for _ in range(1000)]
    >>> sampler = gluoncv.data.SplitSortedBucketSampler(lengths, 16, 1000)
    >>> # The sequence lengths within the batch will be sorted
    >>> for i, indices in enumerate(sampler):
    ...     if i == 0:
    ...         print([lengths[ele] for ele in indices])
    [-etc-]
    """

    def __init__(self, sort_keys, batch_size, mult=32, num_parts=1, part_index=0, shuffle=False,
                 seed=233):
        assert len(sort_keys) > 0
        assert batch_size > 0
        assert mult >= 1, 'Bucket size multiplier must be greater or equal to 1'
        self._sort_keys = sort_keys
        length = len(sort_keys)
        self._batch_size = batch_size
        self._mult = mult
        self._shuffle = shuffle
        # Compute the length of each partition
        part_len = int(np.ceil(float(length) / num_parts))
        # Compute the start index for this partition
        self._start = part_len * part_index
        # Compute the end index for this partition
        self._end = self._start + part_len
        if part_index == num_parts - 1:
            # last part
            self._end = length
            self._start = length - part_len
        self._num_parts = num_parts
        self._seed = seed
        self._shuffled_ids = np.random.RandomState(seed=self._seed).permutation(range(length))

    def __iter__(self):
        if self._num_parts > 1:
            self._shuffled_ids = np.random.RandomState(seed=self._seed).permutation(
                self._shuffled_ids)
        if self._shuffle:
            sample_ids = np.random.permutation(self._shuffled_ids[self._start:self._end])
        else:
            sample_ids = list(range(self._start, self._end))
        bucket_size = int(self._mult * self._batch_size)
        for bucket_begin in range(0, len(sample_ids), bucket_size):
            bucket_end = min(bucket_begin + bucket_size, len(sample_ids))
            if bucket_end - bucket_begin < self._batch_size:
                bucket_begin = bucket_end - self._batch_size
            sorted_sample_ids = sorted(sample_ids[bucket_begin:bucket_end],
                                       key=lambda i: self._sort_keys[i],
                                       reverse=random.randint(0, 1))
            batch_begins = list(range(0, len(sorted_sample_ids), self._batch_size))
            if self._shuffle:
                np.random.shuffle(batch_begins)
            for batch_begin in batch_begins:
                batch_end = min(batch_begin + self._batch_size, len(sorted_sample_ids))
                if batch_end - batch_begin < self._batch_size:
                    yield sorted_sample_ids[batch_end - self._batch_size:batch_end]
                else:
                    yield sorted_sample_ids[batch_begin:batch_end]

    def __len__(self):
        length = int(np.ceil(float(self._end - self._start) / self._batch_size))
        assert length >= 0
        return length



"""
MXNet implementation of SMOT: Single-Shot Multi Object Tracking
https://arxiv.org/abs/2010.16031
"""
from mxnet import gluon
from gluoncv.nn.bbox import BBoxCenterToCorner


class NormalizedLandmarkCenterDecoder(gluon.HybridBlock):
    """
    Decode bounding boxes training target with normalized center offsets.
    This decoder must cooperate with NormalizedBoxCenterEncoder of same `stds`
    in order to get properly reconstructed bounding boxes.

    Returned bounding boxes are using corner type: `x_{min}, y_{min}, x_{max}, y_{max}`.

    Parameters
    ----------
    stds : array-like of size 4
        Std value to be divided from encoded values, default is (0.1, 0.1, 0.2, 0.2).
    means : array-like of size 4
        Mean value to be subtracted from encoded values, default is (0., 0., 0., 0.).
    clip: float, default is None
        If given, bounding box target will be clipped to this value.

    """

    def __init__(self, stds=(0.1, 0.1, 0.2, 0.2), means=(0., 0., 0., 0.),
                 convert_anchor=True):
        super(NormalizedLandmarkCenterDecoder, self).__init__()
        assert len(stds) == 4, "Box Encoder requires 4 std values."
        self._stds = stds
        self._means = means
        if convert_anchor:
            self.center_to_conner = BBoxCenterToCorner(split=True)
        else:
            self.center_to_conner = None

    def hybrid_forward(self, F, x, anchors):
        """center decoder forward"""
        if self.center_to_conner is not None:
            a = self.center_to_conner(anchors)
        else:
            a = anchors.split(axis=-1, num_outputs=4)
        ld = F.split(x, axis=-1, num_outputs=10)

        x0 = F.broadcast_add(F.broadcast_mul(ld[0] * self._stds[0] + self._means[0], a[2] - a[0]), a[0])
        y0 = F.broadcast_add(F.broadcast_mul(ld[1] * self._stds[1] + self._means[1], a[3] - a[1]), a[1])
        x1 = F.broadcast_add(F.broadcast_mul(ld[2] * self._stds[0] + self._means[0], a[2] - a[0]), a[0])
        y1 = F.broadcast_add(F.broadcast_mul(ld[3] * self._stds[1] + self._means[1], a[3] - a[1]), a[1])
        x2 = F.broadcast_add(F.broadcast_mul(ld[4] * self._stds[0] + self._means[0], a[2] - a[0]), a[0])
        y2 = F.broadcast_add(F.broadcast_mul(ld[5] * self._stds[1] + self._means[1], a[3] - a[1]), a[1])
        x3 = F.broadcast_add(F.broadcast_mul(ld[6] * self._stds[0] + self._means[0], a[2] - a[0]), a[0])
        y3 = F.broadcast_add(F.broadcast_mul(ld[7] * self._stds[1] + self._means[1], a[3] - a[1]), a[1])
        x4 = F.broadcast_add(F.broadcast_mul(ld[8] * self._stds[0] + self._means[0], a[2] - a[0]), a[0])
        y4 = F.broadcast_add(F.broadcast_mul(ld[9] * self._stds[1] + self._means[1], a[3] - a[1]), a[1])

        return F.concat(x0, y0, x1, y1, x2, y2, x3, y3, x4, y4, dim=-1)


class GeneralNormalizedKeyPointsDecoder(gluon.HybridBlock):
    """
    Decode bounding boxes training target with normalized center offsets.
    This decoder must cooperate with NormalizedBoxCenterEncoder of same `stds`
    in order to get properly reconstructed bounding boxes.

    Returned bounding boxes are using corner type: `x_{min}, y_{min}, x_{max}, y_{max}`.

    Parameters
    ----------
    stds : array-like of size 4
        Std value to be divided from encoded values, default is (0.1, 0.1, 0.2, 0.2).
    means : array-like of size 4
        Mean value to be subtracted from encoded values, default is (0., 0., 0., 0.).
    clip: float, default is None
        If given, bounding box target will be clipped to this value.

    """

    def __init__(self, num_points, stds=(0.2, 0.2), means=(0.5, 0.2),
                 convert_anchor=True):
        super(GeneralNormalizedKeyPointsDecoder, self).__init__()
        assert len(stds) == 2, "Box Encoder requires 4 std values."
        self._stds = stds
        self._means = means
        self._size = num_points * 2
        if convert_anchor:
            self.center_to_conner = BBoxCenterToCorner(split=True)
        else:
            self.center_to_conner = None

    def hybrid_forward(self, F, x, anchors):
        """key point decoder forward"""
        if self.center_to_conner is not None:
            a = self.center_to_conner(anchors)
        else:
            a = anchors.split(axis=-1, num_outputs=4)
        ld = F.split(x, axis=-1, num_outputs=self._size)

        outputs = []
        for i in range(0, self._size, 2):
            x = F.broadcast_add(F.broadcast_mul(ld[i] * self._stds[0] + self._means[0], a[2] - a[0]), a[0])
            y = F.broadcast_add(F.broadcast_mul(ld[i+1] * self._stds[1] + self._means[1], a[3] - a[1]), a[1])
            outputs.extend([x, y])

        return F.concat(*outputs, dim=-1)



from __future__ import print_function

import numpy as np
import gluoncv as gcv
import mxnet as mx
from mxnet import autograd, gluon
from gluoncv.utils import download, viz

def test_voc07_metric_difficult():
    url = 'https://apache-mxnet.s3-accelerate.amazonaws.com/gluon/dataset/pikachu/train.rec'
    idx_url = 'https://apache-mxnet.s3-accelerate.amazonaws.com/gluon/dataset/pikachu/train.idx'
    download(url, path='pikachu_train.rec', overwrite=False)
    download(idx_url, path='pikachu_train.idx', overwrite=False)
    classes = ['pikachu']

    dataset = gcv.data.RecordFileDetection('pikachu_train.rec')
    net = gcv.model_zoo.get_model('yolo3_darknet53_custom', classes=classes,
                                  pretrained_base=False, transfer='voc')

    def get_dataloader(val_dataset, data_shape, batch_size, num_workers):
        from gluoncv.data.batchify import Tuple, Stack, Pad
        from gluoncv.data.transforms.presets.yolo import YOLO3DefaultValTransform

        val_batchify_fn = Tuple(Stack(), Pad(pad_val=-1))
        val_loader = gluon.data.DataLoader(
            val_dataset.transform(YOLO3DefaultValTransform(data_shape, data_shape)),
            batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep', num_workers=num_workers)
        return val_loader

    val_data = get_dataloader(dataset, 416, 16, 0)

    try:
        a = mx.nd.zeros((1,), ctx=mx.gpu(0))
        ctx = [mx.gpu(0)]
    except:
        ctx = [mx.cpu()]

    from gluoncv.utils.metrics.voc_detection import VOC07MApMetric

    val_metric = VOC07MApMetric(iou_thresh=0.5, class_names=classes)

    def validate(net, val_data, ctx, eval_metric):
        """Test on validation dataset."""
        eval_metric.reset()
        # set nms threshold and topk constraint
        net.set_nms(nms_thresh=0.45, nms_topk=400)
        mx.nd.waitall()
        net.hybridize()
        for batch in val_data:
            data = gluon.utils.split_and_load(batch[0], ctx_list=ctx, batch_axis=0, even_split=False)
            label = gluon.utils.split_and_load(batch[1], ctx_list=ctx, batch_axis=0, even_split=False)
            det_bboxes = []
            det_ids = []
            det_scores = []
            gt_bboxes = []
            gt_ids = []
            gt_difficults = []
            for x, y in zip(data, label):

                # get prediction results
                ids, scores, bboxes = net(x)
                det_ids.append(ids)
                det_scores.append(scores)
                # clip to image size
                det_bboxes.append(bboxes.clip(0, batch[0].shape[2]))
                # split ground truths
                gt_ids.append(y.slice_axis(axis=-1, begin=4, end=5))
                gt_bboxes.append(y.slice_axis(axis=-1, begin=0, end=4))
                gt_difficults.append(y.slice_axis(axis=-1, begin=5, end=6) if y.shape[-1] > 5 else None)

            # update metric
            eval_metric.update(det_bboxes, det_ids, det_scores, gt_bboxes, gt_ids, gt_difficults)
            break
        return eval_metric.get()

    net.collect_params().reset_ctx(ctx)
    validate(net, val_data, ctx, val_metric)


if __name__ == '__main__':
    import nose
    nose.runmodule()



"""Pascal Augmented VOC Semantic Segmentation Dataset."""
import os
import scipy.io
from PIL import Image
from ..segbase import SegmentationDataset

class VOCAugSegmentation(SegmentationDataset):
    """Pascal VOC Augmented Semantic Segmentation Dataset.

    Parameters
    ----------
    root : string
        Path to VOCdevkit folder. Default is '$(HOME)/mxnet/datasplits/voc'
    split: string
        'train' or 'val'
    transform : callable, optional
        A function that transforms the image

    Examples
    --------
    >>> from mxnet.gluon.data.vision import transforms
    >>> # Transforms for Normalization
    >>> input_transform = transforms.Compose([
    >>>     transforms.ToTensor(),
    >>>     transforms.Normalize([.485, .456, .406], [.229, .224, .225]),
    >>> ])
    >>> # Create Dataset
    >>> trainset = gluoncv.data.VOCAugSegmentation(split='train', transform=input_transform)
    >>> # Create Training Loader
    >>> train_data = gluon.data.DataLoader(
    >>>     trainset, 4, shuffle=True, last_batch='rollover',
    >>>     num_workers=4)
    """
    TRAIN_BASE_DIR = 'VOCaug/dataset/'
    NUM_CLASS = 21
    CLASSES = ("background", "airplane", "bicycle", "bird", "boat", "bottle",
               "bus", "car", "cat", "chair", "cow", "diningtable", "dog", "horse",
               "motorcycle", "person", "potted-plant", "sheep", "sofa", "train",
               "tv")

    def __init__(self, root=os.path.expanduser('~/.mxnet/datasets/voc'),
                 split='train', mode=None, transform=None, **kwargs):
        super(VOCAugSegmentation, self).__init__(root, split, mode, transform, **kwargs)
        # train/val/test splits are pre-cut
        _voc_root = os.path.join(root, self.TRAIN_BASE_DIR)
        _mask_dir = os.path.join(_voc_root, 'cls')
        _image_dir = os.path.join(_voc_root, 'img')
        if split == 'train':
            _split_f = os.path.join(_voc_root, 'trainval.txt')
        elif split == 'val':
            _split_f = os.path.join(_voc_root, 'val.txt')
        else:
            raise RuntimeError('Unknown dataset split: {}'.format(split))

        self.images = []
        self.masks = []
        with open(os.path.join(_split_f), "r") as lines:
            for line in lines:
                _image = os.path.join(_image_dir, line.rstrip('\n')+".jpg")
                assert os.path.isfile(_image)
                self.images.append(_image)
                _mask = os.path.join(_mask_dir, line.rstrip('\n')+".mat")
                assert os.path.isfile(_mask)
                self.masks.append(_mask)

        assert (len(self.images) == len(self.masks))

    def __getitem__(self, index):
        img = Image.open(self.images[index]).convert('RGB')
        target = self._load_mat(self.masks[index])
        # synchrosized transform
        if self.mode == 'train':
            img, target = self._sync_transform(img, target)
        elif self.mode == 'val':
            img, target = self._val_sync_transform(img, target)
        else:
            raise RuntimeError('unknown mode for dataloader: {}'.format(self.mode))
        # general resize, normalize and toTensor
        if self.transform is not None:
            img = self.transform(img)
        return img, target

    def _load_mat(self, filename):
        mat = scipy.io.loadmat(filename, mat_dtype=True, squeeze_me=True,
                               struct_as_record=False)
        mask = mat['GTcls'].Segmentation
        return Image.fromarray(mask)

    def __len__(self):
        return len(self.images)

    @property
    def classes(self):
        """Category names."""
        return type(self).CLASSES



from __future__ import print_function

import numpy as np
import mxnet as mx
import gluoncv as gcv
from gluoncv.model_zoo.model_store import pretrained_model_list
from ..unittests.common import try_gpu

@try_gpu(0)
def test_export_model_zoo():
    for model in pretrained_model_list():
        print('exporting:', model)

        if 'deeplab' in model or 'psp' in model or 'icnet' in model:
            # semantic segmentation models require fixed data shape
            kwargs = {'data_shape':(480, 480, 3)}
        elif '3d' in model:
            # video action recognition models require 4d data shape
            kwargs = {'data_shape':(3, 32, 224, 224), 'layout':'CTHW', 'preprocess':None}
        elif 'r2plus1d' in model:
            # video action recognition models require 4d data shape
            kwargs = {'data_shape':(3, 16, 112, 112), 'layout':'CTHW', 'preprocess':None}
        elif 'slowfast_4x16' in model:
            # video action recognition models require 4d data shape
            kwargs = {'data_shape':(3, 36, 224, 224), 'layout':'CTHW', 'preprocess':None}
        elif 'slowfast_8x8' in model:
            # video action recognition models require 4d data shape
            kwargs = {'data_shape':(3, 40, 224, 224), 'layout':'CTHW', 'preprocess':None}
        else:
            kwargs = {}

        if '_gn' in model:
            continue
        if '_dcnv2' in model:
            continue
        if 'siamrpn' in model:
            continue
        if 'danet' in model or 'fastscnn' in model:
            continue
        if 'monodepth2' in model:
            continue

        try:
            gcv.utils.export_block(model, gcv.model_zoo.get_model(model, pretrained=True),
                                   ctx=mx.context.current_context(), **kwargs)
            mx.nd.waitall()
        except ValueError:
            # ignore non defined model name
            pass
        except AttributeError:
            # deeplab model do not support it now, skip
            pass

@try_gpu(0)
def test_export_model_zoo_no_preprocess():
    # special cases
    # 1. no preprocess 2d model
    model_name = 'resnet18_v1b'
    gcv.utils.export_block(model_name, gcv.model_zoo.get_model(model_name, pretrained=True), preprocess=None, layout='CHW')

@try_gpu(0)
def test_rcnn_export_target_generator():
    model_name = 'faster_rcnn_fpn_resnet50_v1b_coco'
    gcv.utils.export_block(model_name, gcv.model_zoo.get_model(model_name, pretrained=True))

def test_export_model_zoo_tvm():
    try:
        import tvm
    except ImportError:
        print("No tvm installed, skipping...")
        return
    for model in ['resnet18_v1']:
        gcv.utils.export_block(model, gcv.model_zoo.get_model(model, pretrained=True))

if __name__ == '__main__':
    import nose
    nose.runmodule()



def normalize(tensor, mean, std):
    """
    Args:
        tensor (Tensor): Tensor to normalize

    Returns:
        Tensor: Normalized tensor
    """
    tensor.sub_(mean).div_(std)
    return tensor



"""Deep Layer Aggregation networks, implemented in Gluon."""
# pylint: disable=arguments-differ,unused-argument,missing-docstring
from __future__ import division

import os

import mxnet as mx
from mxnet.gluon.block import HybridBlock
from mxnet.gluon import nn
from mxnet.gluon.nn import BatchNorm

__all__ = ['DLA', 'get_dla', 'dla34']


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2D(channels=out_planes, kernel_size=3, strides=stride,
                     padding=1, use_bias=False, in_channels=in_planes)


class BasicBlock(HybridBlock):
    def __init__(self, inplanes, planes, stride=1, dilation=1,
                 norm_layer=BatchNorm, norm_kwargs=None, **kwargs):
        super(BasicBlock, self).__init__(**kwargs)
        if norm_kwargs is None:
            norm_kwargs = {}
        with self.name_scope():
            self.conv1 = nn.Conv2D(in_channels=inplanes, channels=planes, kernel_size=3,
                                   strides=stride, padding=dilation,
                                   use_bias=False, dilation=dilation)
            self.bn1 = norm_layer(in_channels=planes, **norm_kwargs)
            self.relu = nn.Activation('relu')
            self.conv2 = nn.Conv2D(in_channels=planes, channels=planes, kernel_size=3,
                                   strides=1, padding=dilation,
                                   use_bias=False, dilation=dilation)
            self.bn2 = norm_layer(in_channels=planes, **norm_kwargs)
            self.stride = stride

    def hybrid_forward(self, F, x, residual=None):
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        out = out + residual
        out = self.relu(out)

        return out


class Bottleneck(HybridBlock):
    expansion = 2

    def __init__(self, inplanes, planes, stride=1, dilation=1,
                 norm_layer=BatchNorm, norm_kwargs=None, **kwargs):
        super(Bottleneck, self).__init__(**kwargs)
        if norm_kwargs is None:
            norm_kwargs = {}
        expansion = Bottleneck.expansion
        bottle_planes = planes // expansion
        with self.name_scope():
            self.conv1 = nn.Conv2D(in_channels=inplanes, channels=bottle_planes,
                                   kernel_size=1, use_bias=False)
            self.bn1 = norm_layer(in_channels=bottle_planes, **norm_kwargs)
            self.conv2 = nn.Conv2D(in_channels=bottle_planes, channels=bottle_planes, kernel_size=3,
                                   strides=stride, padding=dilation,
                                   use_bias=False, dilation=dilation)
            self.bn2 = norm_layer(in_channels=bottle_planes, **norm_kwargs)
            self.conv3 = nn.Conv2D(in_channels=bottle_planes, channels=planes,
                                   kernel_size=1, use_bias=False)
            self.bn3 = norm_layer(**norm_kwargs)
            self.relu = nn.Activation('relu')
            self.stride = stride

    def hybrid_forward(self, F, x, residual=None):
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out = out + residual
        out = self.relu(out)

        return out


class BottleneckX(HybridBlock):
    expansion = 2
    cardinality = 32

    def __init__(self, inplanes, planes, stride=1, dilation=1,
                 norm_layer=BatchNorm, norm_kwargs=None, **kwargs):
        super(BottleneckX, self).__init__(**kwargs)
        if norm_kwargs is None:
            norm_kwargs = {}
        cardinality = BottleneckX.cardinality
        bottle_planes = planes * cardinality // 32
        with self.name_scope():
            self.conv1 = nn.Conv2D(in_channels=inplanes, channels=bottle_planes,
                                   kernel_size=1, use_bias=False)
            self.bn1 = norm_layer(in_channels=bottle_planes, **norm_kwargs)
            self.conv2 = nn.Conv2D(in_channels=bottle_planes, channels=bottle_planes, kernel_size=3,
                                   strides=stride, padding=dilation, use_bias=False,
                                   dilation=dilation, groups=cardinality)
            self.bn2 = norm_layer(in_channels=bottle_planes, **norm_kwargs)
            self.conv3 = nn.Conv2D(in_channels=bottle_planes, channels=planes,
                                   kernel_size=1, use_bias=False)
            self.bn3 = norm_layer(**norm_kwargs)
            self.relu = nn.Activation('relu')
            self.stride = stride

    def hybrid_forward(self, F, x, residual=None):
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out = out + residual
        out = self.relu(out)

        return out


class Root(HybridBlock):
    def __init__(self, in_channels, out_channels, kernel_size, residual,
                 norm_layer=BatchNorm, norm_kwargs=None, **kwargs):
        super(Root, self).__init__(**kwargs)
        if norm_kwargs is None:
            norm_kwargs = {}
        with self.name_scope():
            self.conv = nn.Conv2D(
                in_channels=in_channels, channels=out_channels, kernel_size=1,
                strides=1, use_bias=False, padding=(kernel_size - 1) // 2)
            self.bn = norm_layer(in_channels=out_channels, **norm_kwargs)
            self.relu = nn.Activation('relu')
            self.residual = residual

    def hybrid_forward(self, F, *x):
        children = x
        x = self.conv(F.concat(*x, dim=1))
        x = self.bn(x)
        if self.residual:
            x = x + children[0]
        x = self.relu(x)

        return x


class Tree(HybridBlock):
    def __init__(self, levels, block, in_channels, out_channels, stride=1,
                 level_root=False, root_dim=0, root_kernel_size=1,
                 dilation=1, root_residual=False, norm_layer=BatchNorm, norm_kwargs=None, **kwargs):
        super(Tree, self).__init__(**kwargs)
        if norm_kwargs is None:
            norm_kwargs = {}
        if root_dim == 0:
            root_dim = 2 * out_channels
        if level_root:
            root_dim = root_dim + in_channels
        with self.name_scope():
            self.downsample = nn.HybridSequential()
            self.project = nn.HybridSequential()
            if levels == 1:
                self.tree1 = block(in_channels, out_channels, stride,
                                   dilation=dilation, norm_layer=norm_layer,
                                   norm_kwargs=norm_kwargs, prefix='block_tree_1_')
                self.tree2 = block(out_channels, out_channels, 1,
                                   dilation=dilation, norm_layer=norm_layer,
                                   norm_kwargs=norm_kwargs, prefix='block_tree_2_')
                if in_channels != out_channels:
                    self.project.add(*[
                        nn.Conv2D(in_channels=in_channels, channels=out_channels,
                                  kernel_size=1, strides=1, use_bias=False, prefix='proj_conv0_'),
                        norm_layer(in_channels=out_channels, prefix='proj_bn0_', **norm_kwargs)])
            else:
                self.tree1 = Tree(levels - 1, block, in_channels, out_channels,
                                  stride, root_dim=0,
                                  root_kernel_size=root_kernel_size,
                                  dilation=dilation, root_residual=root_residual,
                                  norm_layer=norm_layer, norm_kwargs=norm_kwargs, prefix='tree_1_')
                self.tree2 = Tree(levels - 1, block, out_channels, out_channels,
                                  root_dim=root_dim + out_channels,
                                  root_kernel_size=root_kernel_size,
                                  dilation=dilation, root_residual=root_residual,
                                  norm_layer=norm_layer, norm_kwargs=norm_kwargs, prefix='tree_2_')
            if levels == 1:
                self.root = Root(root_dim, out_channels, root_kernel_size,
                                 root_residual, norm_layer=norm_layer, norm_kwargs=norm_kwargs,
                                 prefix='root_')
            self.level_root = level_root
            self.root_dim = root_dim
            self.levels = levels
            if stride > 1:
                self.downsample.add(nn.MaxPool2D(stride, strides=stride, prefix='maxpool'))

    def hybrid_forward(self, F, x, residual=None, children=None):
        children = [] if children is None else children
        bottom = self.downsample(x)
        residual = self.project(bottom)
        if self.level_root:
            children.append(bottom)
        x1 = self.tree1(x, residual)
        if self.levels == 1:
            x2 = self.tree2(x1)
            x = self.root(x2, x1, *children)
        else:
            children.append(x1)
            x = self.tree2(x1, None, children)
        return x

class DLA(HybridBlock):
    def __init__(self, levels, channels, classes=1000,
                 block=BasicBlock, momentum=0.9,
                 norm_layer=BatchNorm, norm_kwargs=None,
                 residual_root=False, linear_root=False,
                 use_feature=False, **kwargs):
        super(DLA, self).__init__(**kwargs)
        if norm_kwargs is None:
            norm_kwargs = {}
        norm_kwargs['momentum'] = momentum
        self._use_feature = use_feature
        self.channels = channels
        self.base_layer = nn.HybridSequential('base')
        self.base_layer.add(nn.Conv2D(in_channels=3, channels=channels[0], kernel_size=7, strides=1,
                                      padding=3, use_bias=False))
        self.base_layer.add(norm_layer(in_channels=channels[0], **norm_kwargs))
        self.base_layer.add(nn.Activation('relu'))

        self.level0 = self._make_conv_level(
            channels[0], channels[0], levels[0], norm_layer, norm_kwargs)
        self.level1 = self._make_conv_level(
            channels[0], channels[1], levels[1], norm_layer, norm_kwargs, stride=2)
        self.level2 = Tree(levels[2], block, channels[1], channels[2], 2,
                           level_root=False, root_residual=residual_root,
                           norm_layer=norm_layer, norm_kwargs=norm_kwargs, prefix='level2_')
        self.level3 = Tree(levels[3], block, channels[2], channels[3], 2,
                           level_root=True, root_residual=residual_root,
                           norm_layer=norm_layer, norm_kwargs=norm_kwargs, prefix='level3_')
        self.level4 = Tree(levels[4], block, channels[3], channels[4], 2,
                           level_root=True, root_residual=residual_root,
                           norm_layer=norm_layer, norm_kwargs=norm_kwargs, prefix='level4_')
        self.level5 = Tree(levels[5], block, channels[4], channels[5], 2,
                           level_root=True, root_residual=residual_root,
                           norm_layer=norm_layer, norm_kwargs=norm_kwargs, prefix='level5_')

        if not self._use_feature:
            self.global_avg_pool = nn.GlobalAvgPool2D()
            self.fc = nn.Dense(units=classes)

    def _make_level(self, block, inplanes, planes, blocks, norm_layer, norm_kwargs, stride=1):
        downsample = None
        if stride != 1 or inplanes != planes:
            downsample = nn.HybridSequential()
            downsample.add(*[
                nn.MaxPool2D(stride, strides=stride),
                nn.Conv2D(channels=planes, in_channels=inplanes,
                          kernel_size=1, strides=1, use_bias=False),
                norm_layer(in_channels=planes, **norm_kwargs)])

        layers = []
        layers.append(block(inplanes, planes, stride,
                            norm_layer=norm_layer, norm_kwargs=norm_kwargs, downsample=downsample))
        for _ in range(1, blocks):
            layers.append(block(inplanes, planes, norm_layer=norm_layer, norm_kwargs=norm_kwargs))

        curr_level = nn.HybridSequential()
        curr_level.add(*layers)
        return curr_level

    def _make_conv_level(self, inplanes, planes, convs, norm_layer, norm_kwargs,
                         stride=1, dilation=1):
        modules = []
        for i in range(convs):
            modules.extend([
                nn.Conv2D(in_channels=inplanes, channels=planes, kernel_size=3,
                          strides=stride if i == 0 else 1,
                          padding=dilation, use_bias=False, dilation=dilation),
                norm_layer(**norm_kwargs),
                nn.Activation('relu')])
            inplanes = planes
        curr_level = nn.HybridSequential()
        curr_level.add(*modules)
        return curr_level

    def hybrid_forward(self, F, x):
        y = []
        x = self.base_layer(x)
        for i in range(6):
            x = getattr(self, 'level{}'.format(i))(x)
            if self._use_feature:
                y.append(x)
            else:
                y.append(F.flatten(self.global_avg_pool(x)))
        if self._use_feature:
            return y
        flat = F.concat(*y, dim=1)
        out = self.fc(flat)
        return out

def get_dla(layers, pretrained=False, ctx=mx.cpu(),
            root=os.path.join('~', '.mxnet', 'models'), **kwargs):
    """Get a center net instance.

    Parameters
    ----------
    name : str or int
        Layers of the network.
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    ctx : mxnet.Context
        Context such as mx.cpu(), mx.gpu(0).
    root : str
        Model weights storing path.

    Returns
    -------
    HybridBlock
        A DLA network.

    """
    # pylint: disable=unused-variable
    net = DLA(**kwargs)
    if pretrained:
        from .model_store import get_model_file
        full_name = 'dla{}'.format(layers)
        net.load_parameters(get_model_file(full_name, tag=pretrained, root=root),
                            ctx=ctx, ignore_extra=True)
        from ..data import ImageNet1kAttr
        attrib = ImageNet1kAttr()
        net.synset = attrib.synset
        net.classes = attrib.classes
        net.classes_long = attrib.classes_long
    return net

def dla34(**kwargs):
    """DLA 34 layer network for image classification.

    Returns
    -------
    HybridBlock
        A DLA34 network.

    """
    model = get_dla(34, levels=[1, 1, 1, 2, 2, 1],
                    channels=[16, 32, 64, 128, 256, 512],
                    block=BasicBlock, **kwargs)
    return model



"""Decoder module of Monodepth2
Code partially borrowed from
https://github.com/nianticlabs/monodepth2/blob/master/networks/depth_decoder.py
"""
# Copyright Niantic 2019. Patent Pending. All rights reserved.
#
# This software is licensed under the terms of the Monodepth2 licence
# which allows for non-commercial use only, the full terms of which are made
# available in the LICENSE file.

from __future__ import absolute_import, division, print_function

from collections import OrderedDict
import numpy as np
import mxnet as mx
import mxnet.gluon.nn as nn

from .layers import ConvBlock, Conv3x3


class DepthDecoder(nn.HybridBlock):
    r"""Decoder of Monodepth2

    Parameters
    ----------
    num_ch_enc : list
        The channels number of encoder.
    scales: list
        The scales used in the loss. (Default: range(4))
    num_output_channels: int
        The number of output channels. (Default: 1)
    use_skips: bool
        This will use skip architecture in the network. (Default: True)

    """
    def __init__(self, num_ch_enc, scales=range(4), num_output_channels=1,
                 use_skips=True):
        super(DepthDecoder, self).__init__()

        self.num_output_channels = num_output_channels
        self.use_skips = use_skips
        self.upsample_mode = 'nearest'
        self.scales = scales

        self.num_ch_enc = num_ch_enc
        self.num_ch_dec = np.array([16, 32, 64, 128, 256])

        # decoder
        with self.name_scope():
            self.convs = OrderedDict()
            for i in range(4, -1, -1):
                # upconv_0
                num_ch_in = self.num_ch_enc[-1] if i == 4 else self.num_ch_dec[i + 1]
                num_ch_out = self.num_ch_dec[i]
                self.convs[("upconv", i, 0)] = ConvBlock(num_ch_in, num_ch_out)

                # upconv_1
                num_ch_in = self.num_ch_dec[i]
                if self.use_skips and i > 0:
                    num_ch_in += self.num_ch_enc[i - 1]
                num_ch_out = self.num_ch_dec[i]
                self.convs[("upconv", i, 1)] = ConvBlock(num_ch_in, num_ch_out)

            for s in self.scales:
                self.convs[("dispconv", s)] = Conv3x3(
                    self.num_ch_dec[s], self.num_output_channels)

            # register blocks
            for k in self.convs:
                self.register_child(self.convs[k])
            self.decoder = nn.HybridSequential()
            self.decoder.add(*list(self.convs.values()))

            self.sigmoid = nn.Activation('sigmoid')

    def hybrid_forward(self, F, input_features):
        # pylint: disable=unused-argument, missing-function-docstring
        self.outputs = []

        # decoder
        x = input_features[-1]
        for i in range(4, -1, -1):
            x = self.convs[("upconv", i, 0)](x)
            x = [F.UpSampling(x, scale=2, sample_type='nearest')]
            if self.use_skips and i > 0:
                x += [input_features[i - 1]]
            x = F.concat(*x, dim=1)
            x = self.convs[("upconv", i, 1)](x)
            if i in self.scales:
                self.outputs.append(self.sigmoid(self.convs[("dispconv", i)](x)))

        return self.outputs

    def predict(self, input_features):
        # pylint: disable=unused-argument, missing-function-docstring
        self.outputs = {}

        # decoder
        x = input_features[-1]
        for i in range(4, -1, -1):
            x = self.convs[("upconv", i, 0)].predict(x)
            x = [mx.nd.UpSampling(x, scale=2, sample_type='nearest')]
            if self.use_skips and i > 0:
                x += [input_features[i - 1]]
            x = mx.nd.concat(*x, dim=1)
            x = self.convs[("upconv", i, 1)].predict(x)
            if i in self.scales:
                self.outputs[("disp", i)] = self.sigmoid(self.convs[("dispconv", i)].predict(x))

        return self.outputs



"""Monodepth
Digging Into Self-Supervised Monocular Depth Estimation, ICCV 2019
https://arxiv.org/abs/1806.01260
"""
from mxnet.gluon import nn
from mxnet.context import cpu

from .resnet_encoder import ResnetEncoder
from .pose_decoder import PoseDecoder


class MonoDepth2PoseNet(nn.HybridBlock):
    r"""Monodepth2

    Parameters
    ----------
    backbone : string
        Pre-trained dilated backbone network type ('resnet18', 'resnet34', 'resnet50',
        'resnet101' or 'resnet152').
    pretrained_base : bool or str
        Refers to if the backbone is pretrained or not. If `True`,
        model weights of a model that was trained on ImageNet is loaded.
    num_input_images : int
        The number of input sequences. 1 for depth encoder, larger than 1 for pose encoder.
        (Default: 2)
    num_input_features : int
        The number of input feature maps from posenet encoder. (Default: 1)
    num_frames_to_predict_for: int
        The number of output pose between frames; If None, it equals num_input_features - 1.
        (Default: 2)
    stride: int
        The stride number for Conv in pose decoder. (Default: 1)

    Reference:

        Clement Godard, Oisin Mac Aodha, Michael Firman, Gabriel Brostow.
        "Digging Into Self-Supervised Monocular Depth Estimation." ICCV, 2019

    Examples
    --------
    >>> model = MonoDepth2PoseNet(backbone='resnet18', pretrained_base=True)
    >>> print(model)
    """
    # pylint: disable=unused-argument
    def __init__(self, backbone, pretrained_base, num_input_images=2, num_input_features=1,
                 num_frames_to_predict_for=2, stride=1, ctx=cpu(), **kwargs):
        super(MonoDepth2PoseNet, self).__init__()

        with self.name_scope():
            self.encoder = ResnetEncoder(backbone, pretrained_base,
                                         num_input_images=num_input_images, ctx=ctx)
            if not pretrained_base:
                self.encoder.initialize(ctx=ctx)
            self.decoder = PoseDecoder(self.encoder.num_ch_enc,
                                       num_input_features=num_input_features,
                                       num_frames_to_predict_for=num_frames_to_predict_for,
                                       stride=stride)
            self.decoder.initialize(ctx=ctx)

    def hybrid_forward(self, F, x):
        # pylint: disable=unused-argument
        features = [self.encoder(x)]
        axisangle, translation = self.decoder(features)

        return axisangle, translation

    def demo(self, x):
        return self.predict(x)

    def predict(self, x):
        features = [self.encoder.predict(x)]
        axisangle, translation = self.decoder.predict(features)

        return axisangle, translation


def get_monodepth2posenet(backbone='resnet18', pretrained_base=True, num_input_images=2,
                          num_input_features=1, num_frames_to_predict_for=2, stride=1,
                          root='~/.mxnet/models', ctx=cpu(0), pretrained=False,
                          pretrained_model='kitti_stereo_640x192', **kwargs):
    r"""Monodepth2

    Parameters
    ----------
    backbone : string
        Pre-trained dilated backbone network type ('resnet18', 'resnet34', 'resnet50',
        'resnet101' or 'resnet152').
    pretrained_base : bool or str
        Refers to if the backbone is pretrained or not. If `True`,
        model weights of a model that was trained on ImageNet is loaded.
    num_input_images : int
        The number of input sequences. 1 for depth encoder, larger than 1 for pose encoder.
        (Default: 2)
    num_input_features : int
        The number of input feature maps from posenet encoder. (Default: 1)
    num_frames_to_predict_for: int
        The number of output pose between frames; If None, it equals num_input_features - 1.
        (Default: 2)
    stride: int
        The stride number for Conv in pose decoder. (Default: 1)

    ctx : Context, default: CPU
        The context in which to load the pretrained weights.
    root : str, default: '~/.mxnet/models'
        Location for keeping the model parameters.
    pretrained : bool or str, default: False
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_model : string, default: kitti_stereo_640x192
        The dataset that model pretrained on.

    """

    model = MonoDepth2PoseNet(
        backbone=backbone, pretrained_base=pretrained_base,
        num_input_images=num_input_images, num_input_features=num_input_features,
        num_frames_to_predict_for=num_frames_to_predict_for, stride=stride,
        ctx=ctx, **kwargs)

    if pretrained:
        from ...model_zoo.model_store import get_model_file
        model.load_parameters(
            get_model_file('monodepth2_%s_%s' % (backbone, pretrained_model),
                           tag=pretrained, root=root),
            ctx=ctx
        )
    return model


def get_monodepth2_resnet18_posenet_kitti_mono_640x192(**kwargs):
    r"""Monodepth2 PoseNet

    Parameters
    ----------
    backbone : string
        Pre-trained dilated backbone network type (default:'resnet18').

    """
    return get_monodepth2posenet(backbone='resnet18',
                                 pretrained_model='posenet_kitti_mono_640x192', **kwargs)


def get_monodepth2_resnet18_posenet_kitti_mono_stereo_640x192(**kwargs):
    r"""Monodepth2 PoseNet

    Parameters
    ----------
    backbone : string
        Pre-trained dilated backbone network type (default:'resnet18').

    """
    return get_monodepth2posenet(backbone='resnet18',
                                 pretrained_model='posenet_kitti_mono_stereo_640x192', **kwargs)



import logging
import os
import pickle
import tqdm
from pathlib import Path

from ..dataset import GluonCVMotionDataset, DataSample, FieldNames, SplitNames
from ..utils.ingestion_utils import crop_video, process_dataset_splits, crop_fn_ffmpeg, \
    crop_fn_frame_folder


_log = logging.getLogger()
_log.setLevel(logging.DEBUG)


def even_chunk(sample, chunk_mins, chunks_per_vid=None, center=False, min_chunk_ms=3000):
    """
    A function that when given a sample returns evenly spaced timestamp chunks
    that can be used to crop the video

    :param sample: 
        The datasample to be chunked
    :param chunk_mins: 
        The desired length of each chunk in minutes. 
        This will be the length of all returned chunks unless:
        1. The sample is shorter than chunk_mins OR 
        2. None is provided to chunks_per_vid, then the last chunk might be a shorter length.
    :param chunks_per_vid: 
        The desired number of chunks per video. 
        If a video is too short to be chunked this many times,
        then the number of chunks will be reduced such that they all fit fully. 
        If the video is shorter than a single,
        chunk a single timestamp of the entire video will be returned. 
        If None is provided (the default),
        the video will be chunked fully and the final chunk might be a shorter length
        (but not shorter than min_chunk_ms).
    :param center: 
        Whether to center the chunks (add space before the first and after the last chunk).
        False by default.
    :param min_chunk_ms: 
        Only used when chunks_per_vid is None (chunk the whole video), 
        if the final chunk is smaller than this it will be excluded
    :return: 
        A list of timestamp pairs where each pair is a tuple of (start_time, end_time) for the chunk
    """
    duration = sample.frame_reader.duration
    chunk_ms = int(chunk_mins * 60) * 1000
    if chunks_per_vid is None:
        num_chunks = duration // chunk_ms
        # The final chunk has to be at least a minimum length (or the whole video)
        if duration % chunk_ms >= min_chunk_ms or num_chunks == 0:
            num_chunks += 1
        remaining_ms = 0
    else:
        num_chunks = chunks_per_vid
        remaining_ms = duration - (chunk_ms * num_chunks)
    if remaining_ms < 0:
        # Reduce num chunks by overrun
        # Equivalent of ceil when int dividing with -ve numbers,
        num_chunks += remaining_ms // chunk_ms
        num_chunks = max(1, num_chunks)
        remaining_ms = max(0, duration - (chunk_ms * num_chunks))
    num_spaces = max(1, num_chunks - 1 + (center * 2))
    spacing_ms = remaining_ms // num_spaces
    init_space = spacing_ms if center else 0
    time_starts = [i * (chunk_ms + spacing_ms) + init_space for i in range(num_chunks)]
    time_pairs = [(x, min(x + chunk_ms, duration)) for x in time_starts]
    return time_pairs

def ratio_chunk(sample, ratios):
    """
    :param sample: 
        The datasample to be chunked
    :param ratios: 
        The ratios to be chunked, in order relative to the video.
        If negative it will throw away that chunk
    :return: 
        See :func:`even_chunk`
    """
    if sum(abs(x) for x in ratios) > 1 + 1e-2:
        raise ValueError("Ratios add up to more than 1")

    duration = sample.frame_reader.duration
    time_pairs = []
    start_time = 0
    for ratio in ratios:
        end_time = start_time + (abs(ratio) * duration)
        if ratio > 0:
            time_pair = (round(start_time), round(end_time))
            time_pairs.append(time_pair)
        start_time = end_time
    return time_pairs


def align_times_to_framerate(time_pairs, fps):
    period = 1000 / fps
    aligned_pairs = []
    for time_pair in time_pairs:
        aligned_pair = tuple(round(round(t / period) * period) for t in time_pair)
        aligned_pairs.append(aligned_pair)
    return aligned_pairs


def update_anno(sample: DataSample, new_sample: DataSample, time_pair: (int, int)):
    start_time, end_time = time_pair
    for ts, entities in sample.time_entity_dict.items():
        if ts >= start_time and ts < end_time:
            for entity in entities:
                new_entity = pickle.loads(pickle.dumps(entity))
                new_entity.time = entity.time - start_time
                # This assumes constant fps
                new_entity.frame_num = round((new_entity.time / 1000) * new_sample.fps)
                new_sample.add_entity(new_entity)
    return new_sample


def add_orig_samples(new_dataset, dataset):
    for sid, sample in dataset.get_split_samples("test"):
        new_dataset.add_sample(sample)


def write_new_split(dataset, time_pairs):
    def split_func(sample):
        time_pair = time_pairs.get(sample.id)
        if time_pair is None:
            return SplitNames.TEST
        start_time, end_time = time_pair
        if start_time == 0:
            return SplitNames.TRAIN
        else:
            return SplitNames.VAL

    process_dataset_splits(dataset, split_func, save=True)


def main(anno_path:str="./annotation/anno.json", new_name=None,
         chunk_mins=5, chunks_per_vid=1, chunk_func=even_chunk,
         link_unchanged=True, cache_name=None,input_cache=None,
         name_suffix="", fps=None, split=None, new_split=None,
         overwrite=False, part=0, parts=1):

    fps_suffix = "_{}r".format(fps) if fps is not None else ""
    if name_suffix:
        name_suffix = "_" + name_suffix
    if new_name is None:
        new_name = "anno_chunks_{}m_{}p{}{}.json".format(chunk_mins,
                                                         chunks_per_vid,
                                                         fps_suffix,
                                                         name_suffix)
    if cache_name is None:
        cache_name = "vids_chunked/vid_chunks_{}m_{}p{}{}".format(chunk_mins,
                                                                  chunks_per_vid,
                                                                  fps_suffix,
                                                                  name_suffix)

    dataset = GluonCVMotionDataset(anno_path)
    new_dataset = GluonCVMotionDataset(new_name,
                                       dataset.root_path,
                                       split_file=new_split,
                                       load_anno=False)

    chunk_cache_dir = Path(dataset.cache_root_path, cache_name)
    chunk_data_link = Path(dataset.data_root_path, cache_name)
    os.makedirs(chunk_cache_dir, exist_ok=True)
    os.makedirs(chunk_data_link.parent, exist_ok=True)
    try:
        os.symlink(chunk_cache_dir, chunk_data_link)
    except OSError:
        pass  # already exists

    new_sample_times = {}
    samples = dataset.get_split_samples(split)
    samples = samples[part::parts]
    for sample_id, sample in tqdm.tqdm(samples, mininterval=1):
        time_pairs = ratio_chunk(sample, [0.5, 0.5])
        time_pairs = align_times_to_framerate(time_pairs, sample.fps)
        out_suffix = ".mp4"
        if input_cache is not None:
            input_path = sample.get_cache_file(input_cache, out_suffix)
        else:
            input_path = sample.data_path
        output_path = chunk_data_link / sample.data_relative_path
        is_frame_dir = Path(input_path).is_dir()
        if not is_frame_dir:
            output_path = output_path.with_suffix(out_suffix)

        for start_time, end_time in time_pairs:
            if link_unchanged and len(time_pairs) == 1 \
                    and start_time == 0 and abs(sample.frame_reader.duration - end_time) < 50 \
                    and (fps is None or sample.frame_reader.fps == fps):
                if not output_path.exists():
                    os.makedirs(output_path.parent, exist_ok=True)
                    os.symlink(input_path, output_path)
                cropped = False
                full_output_path = output_path
            else:
                if is_frame_dir:
                    # fps for folder frame reader, not output fps
                    addn_args = dict(crop_fn=crop_fn_frame_folder, in_fps=sample.fps)
                else:
                    # output fps
                    addn_args = dict(crop_fn=crop_fn_ffmpeg, fps=fps)
                cropped, full_output_path = crop_video(input_path, str(output_path),
                                                       start_time, end_time,
                                                       overwrite=overwrite, add_output_ts=True,
                                                       make_dirs=True, **addn_args)
            new_data_path = Path(full_output_path).relative_to(dataset.data_root_path)
            new_id = f"{sample_id}_{start_time}-{end_time}"
            new_sample = sample.get_copy_without_entities(new_id=new_id)
            new_sample.data_relative_path = new_data_path
            fr = new_sample.frame_reader
            new_sample.metadata[FieldNames.NUM_FRAMES] = fr.num_frames()
            new_sample.metadata[FieldNames.DURATION] = fr.duration
            new_sample.metadata[FieldNames.ORIG_ID] = sample_id
            update_anno(sample, new_sample, (start_time, end_time))
            if cropped:
                _log.info("Done crop for {}".format(new_data_path))
            new_dataset.add_sample(new_sample)
            new_sample_times[new_sample.id] = (start_time, end_time)

    if parts == 1:
        add_orig_samples(new_dataset, dataset)
        new_dataset.dump()

    return True, new_dataset


if __name__ == "__main__":
    import fire
    fire.Fire(main)



"""Multigrid support to speed up training of video models"""
from functools import reduce
import numpy as np

from torch.utils.data import Sampler
from torch._six import int_classes as _int_classes


__all__ = ['multiGridSampler', 'MultiGridBatchSampler']


sq2 = np.sqrt(2)
class multiGridSampler(object):
    """
    A Multigrid Method for Efficiently Training Video Models
    Chao-Yuan Wu, Ross Girshick, Kaiming He, Christoph Feichtenhofer, Philipp Krähenbühl
    CVPR 2020, https://arxiv.org/abs/1912.00998
    """
    def __init__(self):
        self.long_cycle = np.asarray([[1, 1, 1], [2, 1, 1], [2, sq2, sq2]])[::-1]
        self.short_cycle = np.asarray([[1, 1, 1], [1, sq2, sq2], [1, 2, 2]])[::-1]
        self.short_cycle_sp = np.asarray([[1, 1, 1], [1, sq2, sq2], [1, sq2, sq2]])[::-1]
        self.mod_long = len(self.long_cycle)
        self.mod_short = len(self.short_cycle)

    def scale(self, x):
        return int(np.around(reduce(lambda a, b: a * b, x)))

    def get_scale(self, alpha, beta):
        long_scale = self.scale(self.long_cycle[alpha])
        if alpha == 0:
            short_scale = self.scale(self.short_cycle_sp[beta])
        else:
            short_scale = self.scale(self.short_cycle[beta])
        return long_scale * short_scale

    def get_scale_alpha(self, alpha):
        long_scale = self.scale(self.long_cycle[alpha])
        return long_scale

    def get_scale_beta(self, beta):
        short_scale = self.scale(self.short_cycle[beta])
        return short_scale

    def get_resize(self, alpha, beta):
        long_item = self.long_cycle[alpha]
        short_item = self.short_cycle[beta]
        return long_item * short_item


class MultiGridBatchSampler(Sampler):
    """
    A Multigrid Method for Efficiently Training Video Models
    Chao-Yuan Wu, Ross Girshick, Kaiming He, Christoph Feichtenhofer, Philipp Krähenbühl
    CVPR 2020, https://arxiv.org/abs/1912.00998
    """
    def __init__(self, sampler, batch_size, drop_last):
        if not isinstance(sampler, Sampler):
            raise ValueError("sampler should be an instance of "
                             "torch.utils.data.Sampler, but got sampler={}"
                             .format(sampler))
        if not isinstance(batch_size, _int_classes) or isinstance(batch_size, bool) or \
                batch_size <= 0:
            raise ValueError("batch_size should be a positive integer value, "
                             "but got batch_size={}".format(batch_size))
        if not isinstance(drop_last, bool):
            raise ValueError("drop_last should be a boolean value, but got "
                             "drop_last={}".format(drop_last))
        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

        self.MG_sampler = multiGridSampler()
        self.alpha = self.MG_sampler.mod_long - 1
        self.beta = 0
        self.batch_scale = self.MG_sampler.get_scale(self.alpha, self.beta)
        self.label = True

    def deactivate(self):
        self.label = False
        self.alpha = self.MG_sampler.mod_long - 1

    def activate(self):
        self.label = True
        self.alpha = 0

    def __iter__(self):
        batch = []
        if self.label:
            self.beta = 0
        else:
            self.beta = self.MG_sampler.mod_short - 1
        self.batch_scale = self.MG_sampler.get_scale(self.alpha, self.beta)
        for idx in self.sampler:
            batch.append([idx, self.alpha, self.beta])
            if len(batch) == self.batch_size*self.batch_scale:
                yield batch
                batch = []
                if self.label:
                    self.beta = (self.beta + 1)%self.MG_sampler.mod_short
                    self.batch_scale = self.MG_sampler.get_scale(self.alpha, self.beta)

        if len(batch) > 0 and not self.drop_last:
            yield batch

    def step_alpha(self):
        self.alpha = (self.alpha + 1)%self.MG_sampler.mod_long

    def compute_lr_milestone(self, lr_milestone):
        """
        long cycle milestones
        """
        self.len_long = self.MG_sampler.mod_long
        self.n_epoch_long = 0
        for x in range(self.len_long):
            self.n_epoch_long += self.MG_sampler.get_scale_alpha(x)
        lr_long_cycle = []
        for i, _ in enumerate(lr_milestone):
            if i == 0:
                pre = 0
            else:
                pre = lr_milestone[i-1]
            cycle_length = (lr_milestone[i] - pre) // self.n_epoch_long
            bonus = (lr_milestone[i] - pre)%self.n_epoch_long // self.len_long
            for j in range(self.len_long)[::-1]:
                pre = pre + cycle_length*(2**j) + bonus
                if j == 0:
                    pre = lr_milestone[i]
                lr_long_cycle.append(pre)
        lr_long_cycle.append(0)
        lr_long_cycle = sorted(lr_long_cycle)
        return lr_long_cycle

    def __len__(self):
        self.len_short = self.MG_sampler.mod_short
        self.n_epoch_short = 0
        for x in range(self.len_short):
            self.n_epoch_short += self.MG_sampler.get_scale_beta(x)
        short_batch_size = self.batch_size * self.MG_sampler.get_scale_alpha(self.alpha)
        num_short = len(self.sampler) // short_batch_size

        total = num_short // self.n_epoch_short * self.len_short
        remain = self.n_epoch_short
        for x in range(self.len_short):
            remain = remain - (2**x)
            if remain <= 0:
                break
            total = total + int(num_short%self.n_epoch_short >= remain)

        return total



"""Estimator implementations"""
from .image_classification import ImageClassificationEstimator
from .ssd import SSDEstimator
from .yolo import YOLOv3Estimator
from .faster_rcnn import FasterRCNNEstimator
# from .mask_rcnn import MaskRCNNEstimator
from .center_net import CenterNetEstimator



"""Monocular Depth Estimation Dataset.
Digging into Self-Supervised Monocular Depth Prediction, ICCV 2019
https://arxiv.org/abs/1806.01260
Code partially borrowed from
https://github.com/nianticlabs/monodepth2/blob/master/datasets/mono_dataset.py
"""
import random
import copy
import numpy as np
from PIL import Image  # using pillow-simd for increased speed

import mxnet as mx
from mxnet.gluon.data import dataset
from mxnet.gluon.data.vision import transforms


def pil_loader(path):
    # open path as file to avoid ResourceWarning
    # (https://github.com/python-pillow/Pillow/issues/835)
    with open(path, 'rb') as f:
        with Image.open(f) as img:
            return img.convert('RGB')


class MonoDataset(dataset.Dataset):
    """Superclass for monocular dataloaders
    Parameters
    ----------
    data_path : string
        Path to dataset folder.
    filenames : string
        Path to split file.
        For example: '$(HOME)/.mxnet/datasets/kitti/splits/eigen_full/train_files.txt'
    height : int
        The height for input images.
    width : int
        The height for input images.
    frame_idxs : list
        The frames to load.
        an integer (e.g. 0, -1, or 1) representing the temporal step relative to 'index',
        or "s" for the opposite image in the stereo pair.
    num_scales : int
        The number of scales of the image relative to the full-size image.
    is_train : bool
        Whether use Data Augmentation. Default is: False
    img_ext : string
        The extension name of input image. Default is '.jpg'
    """

    def __init__(self, data_path, filenames, height, width, frame_idxs,
                 num_scales, is_train=False, img_ext='.jpg'):
        super(MonoDataset, self).__init__()

        self.data_path = data_path
        self.filenames = filenames
        self.height = height
        self.width = width
        self.num_scales = num_scales
        self.interp = Image.ANTIALIAS

        self.frame_idxs = frame_idxs

        self.is_train = is_train
        self.img_ext = img_ext

        self.loader = pil_loader
        self.to_tensor = transforms.ToTensor()

        self.brightness = 0.2
        self.contrast = 0.2
        self.saturation = 0.2
        self.hue = 0.1

        self.load_depth = self.check_depth()

    def preprocess(self, inputs, color_aug):
        """Resize colour images to the required scales and augment if required

        We create the color_aug object in advance and apply the same augmentation to all
        images in this item. This ensures that all images input to the pose network receive the
        same augmentation.
        """
        for k in list(inputs):
            if "color" in k:
                n, im, i = k
                for i in range(self.num_scales):
                    s = 2 ** i
                    size = (self.height // s, self.width // s)
                    inputs[(n, im, i)] = copy.deepcopy(
                        inputs[(n, im, i - 1)].resize(size[::-1], self.interp))

        for k in list(inputs):
            f = mx.nd.array(inputs[k])
            if "color" in k:
                n, im, i = k
                inputs[(n, im, i)] = self.to_tensor(f)
                inputs[(n + "_aug", im, i)] = self.to_tensor(color_aug(f))

    def __len__(self):
        return len(self.filenames)

    def __getitem__(self, index):
        """Returns a single training item from the dataset as a dictionary.

        Values correspond to mxnet NDArray.
        Keys in the dictionary are either strings or tuples:

            ("color", <frame_id>, <scale>)          for raw colour images,
            ("color_aug", <frame_id>, <scale>)      for augmented colour images,
            ("K", scale) or ("inv_K", scale)        for camera intrinsics,
            "stereo_T"                              for camera extrinsics, and
            "depth_gt"                              for ground truth depth maps.

        <frame_id> is either:
            an integer (e.g. 0, -1, or 1) representing the temporal step relative to 'index',
        or
            "s" for the opposite image in the stereo pair.

        <scale> is an integer representing the scale of the image relative to the full-size image:
            -1      images at native resolution as loaded from disk
            0       images resized to (self.width,      self.height     )
            1       images resized to (self.width // 2, self.height // 2)
            2       images resized to (self.width // 4, self.height // 4)
            3       images resized to (self.width // 8, self.height // 8)
        """
        inputs = {}

        do_color_aug = False  # self.is_train and random.random() > 0.5
        do_flip = self.is_train and random.random() > 0.5

        line = self.filenames[index].split()
        folder = line[0]

        if len(line) == 3:
            frame_index = int(line[1])
        else:
            frame_index = 0

        if len(line) == 3:
            side = line[2]
        else:
            side = None

        for i in self.frame_idxs:
            if i == "s":
                other_side = {"r": "l", "l": "r"}[side]
                inputs[("color", i, -1)] = self.get_color(
                    folder, frame_index, other_side, do_flip)
            else:
                inputs[("color", i, -1)] = self.get_color(
                    folder, frame_index + i, side, do_flip)

        # adjusting intrinsics to match each scale in the pyramid
        for scale in range(self.num_scales):
            K = self.K.copy()

            K[0, :] *= self.width // (2 ** scale)
            K[1, :] *= self.height // (2 ** scale)

            inv_K = np.linalg.pinv(K)

            inputs[("K", scale)] = mx.nd.array(K)

            inputs[("inv_K", scale)] = mx.nd.array(inv_K)

        if do_color_aug:
            color_aug = transforms.RandomColorJitter(
                self.brightness, self.contrast, self.saturation, self.hue)
        else:
            color_aug = (lambda x: x)

        self.preprocess(inputs, color_aug)

        for i in self.frame_idxs:
            del inputs[("color", i, -1)]
            del inputs[("color_aug", i, -1)]

        if self.load_depth:
            depth_gt = self.get_depth(folder, frame_index, side, do_flip)
            inputs["depth_gt"] = np.expand_dims(depth_gt, 0)
            inputs["depth_gt"] = mx.nd.array(inputs["depth_gt"].astype(np.float32))

        if "s" in self.frame_idxs:
            stereo_T = np.eye(4, dtype=np.float32)
            baseline_sign = -1 if do_flip else 1
            side_sign = -1 if side == "l" else 1
            stereo_T[0, 3] = side_sign * baseline_sign * 0.1

            inputs["stereo_T"] = mx.nd.array(stereo_T)

        return inputs

    def get_color(self, folder, frame_index, side, do_flip):
        raise NotImplementedError

    def check_depth(self):
        raise NotImplementedError

    def get_depth(self, folder, frame_index, side, do_flip):
        raise NotImplementedError



"""this script is used to help Process dataset for tracking
mainly crop data like_SiamFC"""
import numpy as np
import sys
from ..filesystem import try_import_cv2


def crop_hwc(image, bbox, out_sz, padding=(0, 0, 0)):
    """
    crop image

    Parameters
    ----------
    image: np.array, image
    bbox: np or list, bbox coordinate [xmin,ymin,xmax,ymax]
    out_sz: int , crop image size

    Return:
        crop result
    """
    a = (out_sz-1) / (bbox[2]-bbox[0])
    b = (out_sz-1) / (bbox[3]-bbox[1])
    c = -a * bbox[0]
    d = -b * bbox[1]
    mapping = np.array([[a, 0, c],
                        [0, b, d]]).astype(np.float)
    cv2 = try_import_cv2()
    crop = cv2.warpAffine(image, mapping, (out_sz, out_sz), borderMode=cv2.BORDER_CONSTANT, borderValue=padding)
    return crop

def pos_s_2_bbox(pos, s):
    """
    from center_x,center_y,s to get bbox

    Parameters
    ----------
    pos , x, bbox
    s , int, bbox size

    Return:
        [x_min,y_min,x_max,y_max]
    """
    return [pos[0]-s/2, pos[1]-s/2, pos[0]+s/2, pos[1]+s/2]

def crop_like_SiamFC(image, bbox, context_amount=0.5, exemplar_size=127, instance_size=255, padding=(0, 0, 0)):
    """
    Dataset curation and avoid image resizing during training
    if the tight bounding box has size (w, h) and the context margin is p,
    then the scale factor s is chosen such that the area of the scaled rectangle is equal to a constant
    s(w+2p)×s(h+2p)=A.

    Parameters
    ----------
    image: np.array, image
    bbox: list or np.array, bbox
    context_amount: float, the amount of context to be half of the mean dimension
    exemplar_size: int, exemplar_size
    instance_size: int, instance_size

    Return:
        crop result exemplar z and instance x
    """
    target_pos = [(bbox[2]+bbox[0])/2., (bbox[3]+bbox[1])/2.]
    target_size = [bbox[2]-bbox[0], bbox[3]-bbox[1]]
    wc_z = target_size[1] + context_amount * sum(target_size)
    hc_z = target_size[0] + context_amount * sum(target_size)
    s_z = np.sqrt(wc_z * hc_z)
    scale_z = exemplar_size / s_z
    d_search = (instance_size - exemplar_size) / 2
    pad = d_search / scale_z
    s_x = s_z + 2 * pad

    z = crop_hwc(image, pos_s_2_bbox(target_pos, s_z), exemplar_size, padding)
    x = crop_hwc(image, pos_s_2_bbox(target_pos, s_x), instance_size, padding)
    return z, x

def printProgress(iteration, total, prefix='', suffix='', decimals=1, barLength=100):
    """
    Call in a loop to create terminal progress bar

    Parameters
    ----------
        iteration: int : current iteration.
        total: int, total iterations.
        prefix: str, prefix string.
        suffix: str, suffix string.
        decimals: int, positive number of decimals in percent complete.
        barLength: int, character length of bar.
    """
    formatStr = "{0:." + str(decimals) + "f}"
    percents = formatStr.format(100 * (iteration / float(total)))
    filledLength = int(round(barLength * iteration / float(total)))
    bar = '' * filledLength + '-' * (barLength - filledLength)
    sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percents, '%', suffix)),
    if iteration == total:
        sys.stdout.write('\x1b[2K\r')
    sys.stdout.flush()
